from euphorie.client import model
from euphorie.client import utils
from euphorie.client.docx.compiler import _sanitize_html
from euphorie.client.docx.compiler import DocxCompiler
from euphorie.client.docx.compiler import DocxCompilerFrance
from euphorie.client.docx.compiler import DocxCompilerItaly
from euphorie.client.docx.compiler import DocxCompilerItalyOriginal
from euphorie.client.docx.compiler import DocxCompilerShort
from euphorie.client.docx.compiler import IdentificationReportCompiler
from euphorie.client.utils import get_translated_custom_risks_title
from euphorie.content import MessageFactory as _
from euphorie.content.survey import get_tool_type
from io import BytesIO
from plone import api
from plone.base.utils import safe_text
from plone.memoize.view import memoize
from Products.Five import BrowserView
from sqlalchemy import sql
from urllib.parse import quote
from z3c.saconfig import Session
from zope.i18n import translate


class OfficeDocumentView(BrowserView):
    """Base view that generates an office document and returns it."""

    _compiler = None
    _content_type = ""

    def get_data(self, for_download=False):
        """Return the data for the compiler."""
        return {}

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    def get_payload(self):
        """Compile the template and return the file as a string."""
        output = BytesIO()
        compiler = self._compiler(self.context, self.request)
        compiler.compile(self.get_data(for_download=True))
        compiler.template.save(output)
        output.seek(0)
        return output.read()

    def t(self, txt):
        return translate(txt, context=self.request)

    def get_session_nodes(self):
        """Return an ordered list of all relevant tree items for the current
        survey."""
        query = (
            Session.query(model.SurveyTreeItem)
            .filter(model.SurveyTreeItem.session == self.context.session)
            .filter(sql.not_(model.SKIPPED_PARENTS))
            .filter(
                sql.or_(
                    model.MODULE_WITH_RISK_OR_TOP5_FILTER,
                    model.RISK_PRESENT_OR_TOP5_FILTER,
                )
            )
            .order_by(model.SurveyTreeItem.path)
        )

        return query.all()

    def get_modules(self):
        """Returns the modules for this session."""
        sql_modules = (
            Session.query(model.Module)
            .filter(
                model.SurveyTreeItem.session == self.context.session,
            )
            .order_by(model.SurveyTreeItem.path)
        )
        modules = []
        for sql_module in sql_modules:
            if sql_module.skip_children:
                continue
            if sql_module.zodb_path.find("custom-risks") != -1:
                module_title = get_translated_custom_risks_title(self.request)
            else:
                module_title = sql_module.title
            risks = self.get_risks_for(sql_module)
            modules.append(
                {
                    "title": module_title,
                    "checked": bool(risks),
                    "risks": risks,
                }
            )
        return modules

    def get_risks_for(self, sql_module):
        sql_risks = (
            Session.query(model.Risk)
            .filter(model.Risk.parent_id == sql_module.id)
            .order_by(model.SurveyTreeItem.path)
        )
        risks = []
        for sql_risk in sql_risks:
            if not sql_risk.is_custom_risk:
                risk = self.context.aq_parent.restrictedTraverse(
                    sql_risk.zodb_path.split("/")
                )
                risk_description = risk.description
            else:
                risk = None
                risk_description = ""
            risk_description = _sanitize_html(risk_description)

            # XXX can we move this distinction to the compiler?
            if self._compiler.use_solution_description:
                measures = [
                    f"{risk[measure.solution_id].description}\n{measure.action}"
                    for measure in (sql_risk.in_place_standard_measures)
                ]
            else:
                measures = [
                    measure.action for measure in (sql_risk.in_place_standard_measures)
                ]
            measures.extend(
                [measure.action for measure in (sql_risk.in_place_custom_measures)]
            )

            if sql_risk.identification == "no" or (
                risk and getattr(risk, "type", None) == "top5"
            ):
                actions = [
                    _get_action_plan(
                        action, solution=risk.get(action.solution_id, None)
                    )
                    for action in (sql_risk.standard_measures)
                ]
                actions.extend(
                    [_get_action_plan(action) for action in (sql_risk.custom_measures)]
                )
            else:
                actions = []
            risks.append(
                {
                    "title": sql_risk.title.strip(),
                    "description": risk_description,
                    "comment": sql_risk.comment,
                    "actions": actions,
                    "measures": measures,
                    "epilogue": "",
                    "justifiable": sql_risk.identification,
                    "postponed": sql_risk.postponed,
                    "number": sql_risk.number,
                    "priority": sql_risk.priority,
                    "resources": risk.legal_reference if risk else None,
                }
            )
        return risks

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        self.request.response.setHeader(
            "Content-Disposition",
            f"attachment; filename*=UTF-8''{quote(self._filename)}",
        )
        self.request.response.setHeader(
            "Content-Type",
            self._content_type,
        )
        return self.get_payload()


def _escape_text(txt):
    txt = txt or ""
    txt = txt.replace("<", "&lt;")
    # vertical tab / ASCII 11 / SHIFT + RETURN in Word is unprintable
    txt = txt.replace("\x0b", " ")
    # Also this character can make trouble
    txt = txt.replace("\x01", " ")
    return txt


def _get_action_plan(action, solution=None):
    action_plan = {}
    action_plan["text"] = action.action
    if solution is not None:
        action_plan["solution_description"] = solution.description
    requirements = getattr(action, "requirements") or ""
    action_plan["requirements"] = _escape_text(requirements)
    if action.responsible:
        action_plan["responsible"] = _escape_text(action.responsible)
    if action.planning_start:
        action_plan["planning_start"] = action.planning_start.strftime("%d.%m.%Y")
    if action.planning_end:
        action_plan["planning_end"] = action.planning_end.strftime("%d.%m.%Y")
    return action_plan


class ActionPlanDocxView(OfficeDocumentView):
    """Generate a report based on a basic docx template."""

    _content_type = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"  # noqa: E 501

    @property
    @memoize
    def _compiler(self):
        country = self.webhelpers.country
        tool_type = get_tool_type(self.webhelpers._survey)
        if country == "it":
            if tool_type == "existing_measures":
                return DocxCompilerItaly
            return DocxCompilerItalyOriginal
        if country == "fr":
            if tool_type == "existing_measures":
                return DocxCompilerFrance
        return DocxCompiler

    def get_heading(self, title):
        heading = self.t(
            _(
                "header_oira_report_download",
                default="OiRA Report: “${title}”",
                mapping=dict(title=title),
            ),
        )
        return heading

    def get_section_headings(self):
        headings = [
            self.t(
                _(
                    "header_present_risks",
                    default="Risks that have been identified, "
                    "evaluated and have an Action Plan",
                )
            ),
            self.t(
                _(
                    "header_unevaluated_risks",
                    default="Risks that have been identified but "
                    "do NOT have an Action Plan",
                )
            ),
            self.t(
                _(
                    "header_unanswered_risks",
                    default='Hazards/problems that have been "parked" '
                    "and are still to be dealt with",
                )
            ),
            self.t(
                _(
                    "header_risks_not_present",
                    default="Hazards/problems that have been managed "
                    "or are not present in your organisation",
                )
            ),
        ]
        return headings

    def get_sorted_nodes(self):
        nodes = self.get_session_nodes()
        session = self.context.session

        actioned_nodes = utils.get_actioned_nodes(nodes)
        unactioned_nodes = utils.get_unactioned_nodes(nodes)
        unanswered_nodes = utils.get_unanswered_nodes(session)
        risk_not_present_nodes = utils.get_risk_not_present_nodes(session)
        # From the non-present risks, filter out risks from the (un-)/actioned
        # categories. A "priority" risk will always appear in the action plan,
        # even if it has been answered with "Yes"
        risk_not_present_nodes = [
            n
            for n in risk_not_present_nodes
            if n not in actioned_nodes and n not in unactioned_nodes
        ]
        nodes = [
            actioned_nodes,
            unactioned_nodes,
            unanswered_nodes,
            risk_not_present_nodes,
        ]
        return nodes

    def get_data(self, for_download=False):
        """Gets the data structure in a format suitable for `DocxCompiler`"""
        session = self.context.session
        data = {
            "title": session.title,
            "comment": session.report_comment,
            "heading": self.get_heading(session.title),
            "section_headings": self.get_section_headings(),
            "nodes": self.get_sorted_nodes(),
            "survey_title": self.context.aq_parent.title,
            "modules": self.get_modules(),
        }

        return data

    @property
    def _filename(self):
        """Return the document filename."""
        filename = _(
            "filename_report_actionplan",
            default="Report ${title}",
            mapping={"title": self.context.session.title},
        )
        filename = translate(filename, context=self.request)
        return safe_text(filename) + ".docx"


class ActionPlanShortDocxView(ActionPlanDocxView):
    _compiler = DocxCompilerShort

    @property
    def _filename(self):
        """Return the document filename."""
        filename = _(
            "filename_short_report_actionplan",
            default="Compact report ${title}",
            mapping={"title": self.context.session.title},
        )
        filename = translate(filename, context=self.request)
        return safe_text(filename) + ".docx"


class IdentificationReportDocxView(OfficeDocumentView):
    """Generate a report based on a basic docx template."""

    _compiler = IdentificationReportCompiler
    _content_type = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"  # noqa: E 501

    def get_session_nodes(self):
        """Return an ordered list of all tree items for the current survey.

        By OSHA's request, optional submodules are always included.
        """
        query = (
            Session.query(model.SurveyTreeItem)
            .filter(model.SurveyTreeItem.session == self.context.session)
            .order_by(model.SurveyTreeItem.path)
        )
        return query.all()

    def get_data(self, for_download=False):
        """Gets the data structure in a format suitable for `DocxCompiler`"""

        data = {
            "title": self.context.session.title,
            "heading": "",
            "section_headings": [self.context.session.title],
            "nodes": [self.get_session_nodes()],
        }
        return data

    @property
    def _filename(self):
        """Return the document filename."""
        filename = _(
            "filename_report_identification",
            default="Identification report ${title}",
            mapping=dict(title=self.context.session.title),
        )
        filename = translate(filename, context=self.request)
        return safe_text(filename) + ".docx"



from euphorie.client import model
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.tests.utils import addAccount
from euphorie.client.tests.utils import addSurvey
from euphorie.client.tests.utils import MockMailFixture
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieFunctionalTestCase
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.interfaces import ISecuritySchema
from transaction import commit
from z3c.saconfig import Session
from zExceptions import Unauthorized
from zope.component import getUtility
from zope.interface import alsoProvides

import datetime
import re
import transaction


class GuestAccountTests(EuphorieFunctionalTestCase):
    def test_guest_login_no_valid_survey(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.logout()
        alsoProvides(self.portal.client.REQUEST, IClientSkinLayer)
        browser = self.get_browser()
        api.portal.set_registry_record("euphorie.allow_guest_accounts", True)
        commit()
        browser.open(self.portal.client.nl.absolute_url())
        self.assertIn("Start a test session", browser.contents)
        # No valid survey path is passed in came_from
        browser.open(
            "%s/@@surveys?came_from=%s"
            % (
                self.portal.client.nl.absolute_url(),
                self.portal.client.nl.absolute_url(),
            )
        )
        # Therefore we land on the "start new session" page
        self.assertIn("Test session", browser.contents)

    def test_guest_login_with_valid_survey(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.logout()
        alsoProvides(self.portal.client.REQUEST, IClientSkinLayer)
        browser = self.get_browser()
        api.portal.set_registry_record("euphorie.allow_guest_accounts", True)
        commit()
        browser.open(self.portal.client.nl.absolute_url())
        self.assertIn("Start a test session", browser.contents)
        url = f"{self.portal.client.nl.absolute_url()}/ict/software-development"
        # We pass in a valid survey path in came_from
        browser.open("{url}/@@tryout?came_from={url}".format(url=url))
        # Therefore we land on the start page of the survey
        self.assertIn("Test session", browser.contents)
        self.assertIn("<h1>Software development</h1>", browser.contents)


class LoginTests(EuphorieFunctionalTestCase):
    def test_login_not_case_sensitive(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        addAccount(password="secret")
        browser = self.get_browser()
        browser.open(self.portal.client.nl.absolute_url() + "/@@login")
        browser.getControl(name="__ac_name").value = "JANE@example.com"
        browser.getControl(name="__ac_password:utf8:ustring").value = "secret"
        browser.getControl(name="login").click()
        self.assertTrue("@@login" not in browser.url)

    def test_use_session_cookie_by_default(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        addAccount(password="secret")
        browser = self.get_browser()
        browser.open(self.portal.client.nl.absolute_url() + "/@@login")
        browser.getControl(name="__ac_name").value = "jane@example.com"
        browser.getControl(name="__ac_password:utf8:ustring").value = "secret"
        browser.getControl(name="login").click()
        auth_cookie = browser.cookies.getinfo("__ac")
        self.assertEqual(auth_cookie["expires"], None)

    def test_remember_user_sets_cookie_expiration(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        addAccount(password="secret")
        browser = self.get_browser()
        browser.open(self.portal.client.nl.absolute_url() + "/@@login")
        browser.getControl(name="__ac_name").value = "jane@example.com"
        browser.getControl(name="__ac_password:utf8:ustring").value = "secret"
        browser.getControl(name="remember").value = True
        browser.getControl(name="login").click()
        auth_cookie = browser.cookies.getinfo("__ac")
        self.assertNotEqual(auth_cookie["expires"], None)
        delta = auth_cookie["expires"] - datetime.datetime.now(
            auth_cookie["expires"].tzinfo
        )
        self.assertTrue(delta.days > 100)

    def test_extra_ga_pageview_post_login(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        addAccount(password="secret")
        browser = self.get_browser()
        browser.open(self.portal.client.nl.absolute_url() + "/@@login")
        browser.getControl(name="__ac_name").value = "JANE@example.com"
        browser.getControl(name="__ac_password:utf8:ustring").value = "secret"
        browser.getControl(name="login").click()
        self.assertTrue(
            re.search("trackPageview.*login/success", browser.contents) is not None
        )

    def test_record_last_login_time(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        account = addAccount(password="secret")
        browser = self.get_browser()
        browser.open(self.portal.client.nl.absolute_url() + "/@@login")
        browser.getControl(name="__ac_name").value = "jane@example.com"
        browser.getControl(name="__ac_password:utf8:ustring").value = "secret"
        browser.getControl(name="login").click()
        last_login = Session().query(account.__class__).one().last_login
        delta = datetime.datetime.now(last_login.tzinfo) - last_login
        self.assertAlmostEqual(delta.seconds / 10, 0)


class RegisterTests(EuphorieIntegrationTestCase):
    def test_lowercase_email(self):
        with self._get_view("login", self.portal.client) as view:
            view.errors = {}
            view.request.form["email"] = "JANE@example.com"
            view.request.form["password1"] = "Secret123Secret"
            view.request.form["password2"] = "Secret123Secret"
            view.request.form["terms"] = "on"
            account = view._tryRegistration()
            self.assertEqual(account.loginname, "jane@example.com")

    def test_first_name_last_name(self):
        with self._get_view("login", self.portal.client) as view:
            view.errors = {}
            view.request.form["first_name"] = "Jane"
            view.request.form["last_name"] = "Doe"
            view.request.form["email"] = "JANE@example.com"
            view.request.form["password1"] = "Secret123Secret"
            view.request.form["password2"] = "Secret123Secret"
            view.request.form["terms"] = "on"
            account = view._tryRegistration()
            self.assertEqual(account.loginname, "jane@example.com")
            self.assertEqual(account.first_name, "Jane")
            self.assertEqual(account.last_name, "Doe")

    def test_conflict_with_plone_account(self):
        with self._get_view("login", self.portal.client) as view:
            view.errors = {}
            view.request.form["email"] = self.portal._owner[1]
            view.request.form["password1"] = "Secret123Secret"
            view.request.form["password2"] = "Secret123Secret"
            view.request.form["terms"] = "on"
            self.assertEqual(view._tryRegistration(), False)
            self.assertTrue("email" in view.errors)

    def test_basic_email_verification(self):
        with self._get_view("login", self.portal.client) as view:
            view.errors = {}
            view.request.form["email"] = "wichert"
            view.request.form["password1"] = "Secret123Secret"
            view.request.form["password2"] = "Secret123Secret"
            view.request.form["terms"] = "on"
            self.assertEqual(view._tryRegistration(), False)
            self.assertTrue("email" in view.errors)

            view.errors.clear()
            view.request.form["email"] = "wichert@wiggy net"
            self.assertEqual(view._tryRegistration(), False)
            self.assertTrue("email" in view.errors)

            view.errors.clear()
            view.request.form["email"] = "wichert@wiggy.net"
            self.assertNotEqual(view._tryRegistration(), False)

    def test_terms_not_accepted(self):
        with self._get_view("login", self.portal.client) as view:
            view.errors = {}
            view.request.form["email"] = self.portal._owner[1]
            view.request.form["password1"] = "Secret123Secret"
            view.request.form["password2"] = "Secret123Secret"
            self.assertEqual(view._tryRegistration(), False)
            self.assertTrue("terms" in view.errors)

    def test_registration_not_allowed(self):
        registry = getUtility(IRegistry)
        security_settings = registry.forInterface(ISecuritySchema, prefix="plone")
        security_settings.enable_self_reg = False
        with self._get_view("login", self.portal.client) as view:
            view.errors = {}
            view.request.form["email"] = "jane@example.com"
            view.request.form["password1"] = "Secret123Secret"
            view.request.form["password2"] = "Secret123Secret"
            view.request.form["terms"] = "on"
            with self.assertRaises(Unauthorized):
                view._tryRegistration()


class ResetPasswordTests(EuphorieFunctionalTestCase):
    def add_dummy_survey(self):
        survey = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
                      <title>Sector title</title>
                      <survey>
                        <title>Survey title</title>
                      </survey>
                    </sector>"""
        self.loginAsPortalOwner()
        addSurvey(self.portal, survey)
        self.logout()

    def test_unknown_account(self):
        self.add_dummy_survey()
        browser = self.get_browser()
        url = self.portal.client.nl.absolute_url()

        browser.open(url + "/@@login")
        browser.getLink("I forgot my password").click()
        browser.getControl(name="form.widgets.email").value = "jane@example.com"
        browser.getControl(name="form.buttons.save").click()

        # We do not have any account here
        self.assertListEqual(model.Session.query(model.Account).all(), [])
        # Even if the user does not exist, the form submission
        # will be successfully sent
        self.assertTrue(
            browser.url.startswith("http://nohost/plone/client/nl/@@login?came_from")
        )

    def test_invalid_email(self):
        self.add_dummy_survey()
        browser = self.get_browser()
        url = self.portal.client.nl.absolute_url()

        browser.open(url + "/@@login")
        browser.getLink("I forgot my password").click()
        # Test an invalid email address
        browser.getControl(name="form.widgets.email").value = "jane @ joe.com"
        browser.getControl(name="form.buttons.save").click()
        self.assertIn("The specified email is not valid.", browser.contents)
        self.assertEqual(
            browser.url,
            "http://nohost/plone/client/nl/@@reset_password_request",
        )

    def test_email(self):
        self.add_dummy_survey()
        addAccount()
        mail_fixture = MockMailFixture()
        browser = self.get_browser()
        browser.open(self.portal.client.nl.absolute_url() + "/@@login")
        browser.getLink("I forgot my password").click()
        browser.getControl(name="form.widgets.email").value = "jane@example.com"
        browser.getControl(name="form.buttons.save").click()
        self.assertEqual(len(mail_fixture.storage), 1)
        (args, kw) = mail_fixture.storage[0]
        (mail, mto, mfrom) = args[:3]
        self.assertEqual(mfrom, "discard@simplon.biz")
        self.assertEqual(mto, "jane@example.com")
        self.assertEqual(
            str(mail["Subject"]),
            "OiRA password reset instructions",
        )
        body = (
            mail.get_payload(0)
            .get_payload(decode=True)
            .decode(mail.get_content_charset("utf-8"))
        )
        self.assertTrue("/passwordreset/" in body)

    def test_invalid_reset_key(self):
        """When the request key is invalid the user is invited to request a new
        key."""
        self.add_dummy_survey()
        browser = self.get_browser()
        for url in (
            self.portal.client.nl.absolute_url() + "/passwordreset",
            self.portal.client.nl.absolute_url() + "/passwordreset/foo",
        ):
            browser.open(url)
            browser.getControl(name="form.widgets.new_password").value = (
                "Secret123Secret"
            )
            browser.getControl(name="form.widgets.new_password_confirmation").value = (
                "Secret123Secret"
            )
            browser.getControl(label="Save changes").click()
            self.assertIn("Invalid security token", browser.contents)

    def test_token_invalid_after_use(self):
        self.add_dummy_survey()
        addAccount()
        mail_fixture = MockMailFixture()

        browser = self.get_browser()
        url = self.portal.client.nl.absolute_url()

        browser.open(url + "/@@login")
        browser.getLink("I forgot my password").click()
        browser.getControl(name="form.widgets.email").value = "jane@example.com"
        browser.getControl(name="form.buttons.save").click()

        args = mail_fixture.storage[0][0]
        mail = args[0]
        mail_text = "".join(
            [
                (part.get_payload(decode=True) or b"").decode(
                    part.get_content_charset("iso-8859-1")
                )
                for part in mail.walk()
            ]
        )

        reset_url = re.search("http.*passwordreset/\\S*", mail_text).group(0)
        browser.open(reset_url)
        browser.getControl(name="form.widgets.new_password").value = "Test12345678"
        browser.getControl(name="form.widgets.new_password_confirmation").value = (
            "Test12345678"
        )
        browser.getControl(name="form.buttons.save").click()

        self.assertIn("success", browser.contents)

        # Token has been used already - second time should fail
        browser.open(reset_url)
        self.assertIn("Invalid security token", browser.contents)

        # You're free to fill in the form but it won't work
        browser.getControl(name="form.widgets.new_password").value = "Test12345670"
        browser.getControl(name="form.widgets.new_password_confirmation").value = (
            "Test12345670"
        )
        browser.getControl(name="form.buttons.save").click()

        self.assertNotIn("success", browser.contents)

    def test_token_invalid_after_new_request(self):
        self.add_dummy_survey()
        addAccount()
        mail_fixture = MockMailFixture()

        browser = self.get_browser()
        url = self.portal.client.nl.absolute_url()

        browser.open(url + "/@@login")
        browser.getLink("I forgot my password").click()
        browser.getControl(name="form.widgets.email").value = "jane@example.com"
        browser.getControl(name="form.buttons.save").click()

        # Request another password reset without using the first one

        browser.open(url + "/@@login")
        browser.getLink("I forgot my password").click()
        browser.getControl(name="form.widgets.email").value = "jane@example.com"
        browser.getControl(name="form.buttons.save").click()

        args = mail_fixture.storage[0][0]
        mail = args[0]
        mail_text = "".join(
            [
                (part.get_payload(decode=True) or b"").decode(
                    part.get_content_charset("iso-8859-1")
                )
                for part in mail.walk()
            ]
        )

        # Now try using the first token
        reset_url = re.search("http.*passwordreset/\\S*", mail_text).group(0)
        browser.open(reset_url)
        self.assertIn("Invalid security token", browser.contents)

        # You're free to fill in the form but it won't work
        browser.getControl(name="form.widgets.new_password").value = "Test12345678"
        browser.getControl(name="form.widgets.new_password_confirmation").value = (
            "Test12345678"
        )
        browser.getControl(name="form.buttons.save").click()

        self.assertNotIn("success", browser.contents)

    def test_token_expired(self):
        self.add_dummy_survey()
        addAccount()
        mail_fixture = MockMailFixture()

        browser = self.get_browser()
        url = self.portal.client.nl.absolute_url()

        browser.open(url + "/@@login")
        browser.getLink("I forgot my password").click()
        browser.getControl(name="form.widgets.email").value = "jane@example.com"
        browser.getControl(name="form.buttons.save").click()

        args = mail_fixture.storage[0][0]
        mail = args[0]
        mail_text = "".join(
            [
                (part.get_payload(decode=True) or b"").decode(
                    part.get_content_charset("iso-8859-1")
                )
                for part in mail.walk()
            ]
        )
        token = re.search("passwordreset/(\\S*)", mail_text).group(1)
        # fake that the token has expired
        ppr = api.portal.get_tool("portal_password_reset")
        ppr._requests[token] = (ppr._requests[token][0], datetime.datetime(2001, 1, 1))
        ppr._p_changed = 1
        transaction.commit()

        reset_url = re.search("http.*passwordreset/\\S*", mail_text).group(0)
        browser.open(reset_url)
        self.assertIn("Invalid security token", browser.contents)

        # You're free to fill in the form but it won't work

        browser.getControl(name="form.widgets.new_password").value = "Test12345678"
        browser.getControl(name="form.widgets.new_password_confirmation").value = (
            "Test12345678"
        )
        browser.getControl(name="form.buttons.save").click()

        self.assertNotIn("success", browser.contents)
        self.assertIn("Invalid security token", browser.contents)

    def test_token_expires_after_12_hours(self):
        self.add_dummy_survey()
        addAccount()
        mail_fixture = MockMailFixture()

        browser = self.get_browser()
        url = self.portal.client.nl.absolute_url()

        browser.open(url + "/@@login")
        browser.getLink("I forgot my password").click()
        browser.getControl(name="form.widgets.email").value = "jane@example.com"
        browser.getControl(name="form.buttons.save").click()

        args = mail_fixture.storage[0][0]
        mail = args[0]
        mail_text = "".join(
            [
                (part.get_payload(decode=True) or b"").decode(
                    part.get_content_charset("iso-8859-1")
                )
                for part in mail.walk()
            ]
        )
        token = re.search("passwordreset/(\\S*)", mail_text).group(1)
        ppr = api.portal.get_tool("portal_password_reset")
        _, expiry = ppr._requests[token]
        self.assertLessEqual(
            expiry, datetime.datetime.now() + datetime.timedelta(hours=12)
        )



from plone.app.imagecropping.browser.settings import ISettings
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.utils import getAllowedSizes
from zope.component import getMultiAdapter
from zope.component._api import getUtility
from zope.globalrequest import getRequest


FIELDNAMES = ("image", "external_site_logo")


def _initial_size(ix, iy, sx, sy):
    """we need a best fit centered selection.

    Shamelessly copied from plone.app.imaging.browser.editor
    """
    ix, iy, sx, sy = map(float, (ix, iy, sx, sy))
    # aspect ratio of original
    if iy > 0:
        ir = ix / iy
    else:
        ir = 1

    # aspect ratio of scale
    if sy > 0:
        sr = sx / sy
    else:
        sr = 1

    # scale up to bounds
    if ir > sr:
        rx1, ry1 = ix * sr / ir, iy
    else:
        rx1, ry1 = ix, iy * ir / sr

    rx0, ry0 = 0, 0

    # center box
    if rx1 < ix:
        deltax = ix - rx1
        rx0 = deltax / 2
        rx1 = rx1 + deltax / 2
    if ry1 < iy:
        deltay = iy - ry1
        ry0 = deltay / 2
        ry1 = ry1 + deltay / 2

    # round to int
    rx0, ry0 = int(round(rx0)), int(round(ry0))
    rx1, ry1 = int(round(rx1)), int(round(ry1))
    return rx0, ry0, rx1, ry1


def _autocrop_scales(context):
    cropping_registry = getUtility(IRegistry)
    settings = cropping_registry.forInterface(ISettings)
    scale_names = settings.cropping_for
    request = getRequest()
    cropper = getMultiAdapter((context, request), name="crop-image")
    for fname in FIELDNAMES:
        field = getattr(context, fname, None)
        if not field:
            continue

        real_width, real_height = map(float, field.getImageSize())
        # some images are not initialized properly, in which case they
        # get -1 for width and height. In this case, don't crop
        if real_width == -1 or real_height == -1:
            continue
        allowed = getAllowedSizes()
        for scale_name in scale_names:
            if scale_name not in allowed:
                continue

            width, height = map(float, allowed[scale_name])

            coords = _initial_size(real_width, real_height, width, height)
            cropper._crop(fname, scale_name, coords)


def crop_on_image_edit(context, event):
    _autocrop_scales(context)



"""Empty message.

Revision ID: 20231019121221
Revises: 20230515143459
Create Date: 2023-10-19 17:58:47.014626
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_table

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20231019121221"
down_revision = "20230515143459"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("notification_subscription"):
        op.create_table(
            "notification_subscription",
            sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
            sa.Column("account_id", sa.Integer(), nullable=False),
            sa.Column("category", sa.String(length=512), nullable=False),
            sa.Column("enabled", sa.Boolean(), nullable=False, default=False),
            sa.ForeignKeyConstraint(
                ["account_id"], ["account.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )


def downgrade():
    op.drop_table("notification_subscription")



from euphorie.client.model import SurveySession
from euphorie.client.model import Training
from plone import api
from plone.memoize.view import memoize
from Products.Five import BrowserView
from z3c.saconfig import Session


class Certificate(BrowserView):
    @property
    @memoize
    def trainings(self):
        """Get all trainings for this session."""
        session_id = self.context.session.id
        return (
            Session.query(Training)
            .filter(Training.session_id == session_id, Training.status == "correct")
            .order_by(Training.time.desc())
            .all()
        )

    def get_certificate_info(self, training):
        traversed_session = training.session.traversed_session
        certificate_view = api.content.get_view(
            name="training-certificate-inner",
            context=traversed_session,
            request=self.request,
        )
        content = certificate_view.index(training_id=training.id)
        link = (
            f"{training.session.absolute_url()}/@@training-certificate-view"
            f"?training_id={training.id}"
        )
        return {"link": link, "content": content, "date": training.time}

    @property
    @memoize
    def certificates(self):
        certificates = []
        for training in self.trainings:
            certificates.append(self.get_certificate_info(training))
        return certificates


class CertificateOverview(Certificate):
    """Certificates Overview Page"""

    @property
    @memoize
    def trainings(self):
        """Get all trainings from all the current user's organisations."""
        organisation_view = api.content.get_view(
            name="organisation",
            context=self.context,
            request=self.request,
        )
        account_ids = {
            organisation.owner_id for organisation in organisation_view.organisations
        }
        account_ids.add(api.user.get_current().getId())
        return [
            training
            for training in (
                Session.query(Training)
                .filter(
                    Training.session_id == SurveySession.id,
                    SurveySession.account_id.in_(account_ids),
                )
                .filter(Training.status == "correct")
                .order_by(Training.time.desc())
                .all()
            )
            if training.session.tool
        ]

    @property
    @memoize
    def certificates(self):
        """Get all certificates that I am allowed to view, grouped by year."""
        certificates = {}
        for training in self.trainings:
            year = training.time.year
            certificates.setdefault(year, []).append(
                self.get_certificate_info(training)
            )
        for year, year_certificates in certificates.items():
            certificates[year] = sorted(
                year_certificates, key=lambda c: c["date"], reverse=True
            )
        return certificates.items()

    def get_num_columns(self, num_items):
        return 2 if num_items < 3 else 3



"""Try to add an image binary column and an image_filename column to the risk
table Also add column "refreshed" to the session table and copy data from
modified.

Revision ID: 24
Revises:
Create Date: 2019-09-09 10:24:51.539339
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "24"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("risk", "image_data"):
        op.add_column("risk", sa.Column("image_data", sa.LargeBinary(), nullable=True))
    if not has_column("risk", "image_data_scaled"):
        op.add_column(
            "risk", sa.Column("image_data_scaled", sa.LargeBinary(), nullable=True)
        )
    if not has_column("risk", "image_filename"):
        op.add_column(
            "risk", sa.Column("image_filename", sa.UnicodeText(), nullable=True)
        )
    if not has_column("session", "refreshed"):
        op.add_column("session", sa.Column("refreshed", sa.DateTime(), nullable=True))
        op.execute("UPDATE session SET refreshed = modified")


def downgrade():
    op.drop_column("risk", "image_data")
    op.drop_column("risk", "image_data_scaled")
    op.drop_column("risk", "image_filename")
    op.drop_column("session", "refreshed")



"""
Module
------

A module is a grouping for risks in a survey. It can be set to optional, in
which case all risks in it can be skipped. A module can contain other modules.

portal_type: euphorie.module
"""

from .. import MessageFactory as _
from .behaviour.richdescription import IRichDescription
from .behaviour.uniqueid import get_next_id
from .behaviour.uniqueid import INameFromUniqueId
from .fti import check_fti_paste_allowed
from .fti import ConditionalDexterityFTI
from .fti import IConstructionFilter
from .interfaces import IQuestionContainer
from .utils import StripMarkup
from Acquisition import aq_chain
from euphorie.content.dependency import ConditionalTextLine
from euphorie.content.utils import ensure_image_size
from euphorie.htmllaundry.z3cform import HtmlText
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.autoform import directives
from plone.dexterity.content import Container
from plone.indexer import indexer
from plone.namedfile import field as filefield
from plone.namedfile.interfaces import INamedBlobImageField
from plone.supermodel import model
from plonetheme.nuplone.z3cform.directives import depends
from plonetheme.nuplone.z3cform.widget import WysiwygFieldWidget
from Products.statusmessages.interfaces import IStatusMessage
from z3c.form import validator
from zope import schema
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import Invalid

import sys


class IModule(model.Schema, IRichDescription, IBasic):
    """Survey Module.

    A module is (hierarchical) grouping in a survey.
    """

    description = HtmlText(
        title=_("label_module_description", "Description"),
        description=_(
            "help_module_description",
            default="Include any relevant information that may be "
            "helpful for the end-user.",
        ),
        required=False,
    )
    directives.widget(description=WysiwygFieldWidget)
    directives.order_after(description="title")

    optional = schema.Bool(
        title=_("label_module_optional", default="This module is optional"),
        description=_(
            "help_module_optional",
            default="Allows the end-user to skip this module and "
            "everything inside it.",
        ),
        required=False,
        default=False,
    )

    depends("question", "optional", "on")
    question = ConditionalTextLine(
        title=_("label_module_question", default="Question"),
        description=_(
            "help_module_question",
            default="The question to ask the end-user if this module is "
            "optional. It must be formulated so that it is answerable "
            "with YES (the end-user will have to tick a box) or NO",
        ),
        required=True,
    )

    image = filefield.NamedBlobImage(
        title=_("label_image", default="Image file"),
        description=_(
            "help_image_upload",
            default="Upload an image. Make sure your image is of format "
            "png, jpg or gif and does not contain any special "
            "characters. The minimum size is 1000 (width) x 430 (height) pixels.",
        ),
        required=False,
    )
    caption = schema.TextLine(
        title=_("label_caption", default="Image caption"), required=False
    )

    solution_direction = HtmlText(
        title=_("label_solution_direction", default="Solution"),
        description=_(
            "help_solution_direction",
            default="This information will appear in the Action plan step "
            "and should include an overview of general solution(s) "
            "related to this module.",
        ),
        required=False,
    )
    directives.widget(solution_direction=WysiwygFieldWidget)

    model.fieldset(
        "additional_content",
        label=_("header_additional_content", default="Additional content"),
        description=_(
            "intro_additional_content",
            default="Attach any additional content you consider helpful "
            "for the user",
        ),
        fields=[
            "file1",
            "file1_caption",
            "file2",
            "file2_caption",
            "file3",
            "file3_caption",
            "file4",
            "file4_caption",
        ],
    )

    file1 = filefield.NamedBlobFile(
        title=_("label_file", default="Content file"),
        description=_(
            "help_content_upload",
            default="Upload a file that contains additional information, "
            "like a PDF, Word document or spreadsheet. Optionally provide "
            "a descriptive caption for your file.",
        ),
        required=False,
    )
    file1_caption = schema.TextLine(
        title=_("label_file_caption", default="Content caption"), required=False
    )

    file2 = filefield.NamedBlobFile(
        title=_("label_file", default="Content file"),
        description=_(
            "help_content_upload",
            default="Upload a file that contains additional information, "
            "like a PDF, Word document or spreadsheet. Optionally provide "
            "a descriptive caption for your file.",
        ),
        required=False,
    )
    file2_caption = schema.TextLine(
        title=_("label_file_caption", default="Content caption"), required=False
    )

    file3 = filefield.NamedBlobFile(
        title=_("label_file", default="Content file"),
        description=_(
            "help_content_upload",
            default="Upload a file that contains additional information, "
            "like a PDF, Word document or spreadsheet. Optionally provide "
            "a descriptive caption for your file.",
        ),
        required=False,
    )
    file3_caption = schema.TextLine(
        title=_("label_file_caption", default="Content caption"), required=False
    )

    file4 = filefield.NamedBlobFile(
        title=_("label_file", default="Content file"),
        description=_(
            "help_content_upload",
            default="Upload a file that contains additional information, "
            "like a PDF, Word document or spreadsheet. Optionally provide "
            "a descriptive caption for your file.",
        ),
        required=False,
    )
    file4_caption = schema.TextLine(
        title=_("label_file_caption", default="Content caption"), required=False
    )


class ImageSizeValidator(validator.SimpleFieldValidator):
    def validate(self, value):
        try:
            previous_value = self.field.get(self.context)
        except AttributeError:
            previous_value = None
        if previous_value == value:
            try:
                ensure_image_size(value)
            except Invalid as invalid:
                IStatusMessage(self.context.REQUEST).add(invalid, "warn")
        else:
            ensure_image_size(value)


validator.WidgetValidatorDiscriminators(
    ImageSizeValidator, context=IModule, field=INamedBlobImageField
)


@implementer(IModule, IQuestionContainer)
class Module(Container):
    image = None
    caption = None
    file1 = None
    file1_caption = None
    file2 = None
    file2_caption = None
    file3 = None
    file3_caption = None
    file4 = None
    file4_caption = None

    def _get_id(self, orig_id):
        """Pick an id for pasted content."""
        frame = sys._getframe(1)
        ob = frame.f_locals.get("ob")
        if ob is not None and INameFromUniqueId.providedBy(ob):
            return get_next_id(self)
        return super()._get_id(orig_id)

    def _verifyObjectPaste(self, object, validate_src=True):
        super()._verifyObjectPaste(object, validate_src)
        if validate_src:
            check_fti_paste_allowed(self, object)
            if IQuestionContainer.providedBy(object):
                my_depth = item_depth(self)
                paste_depth = tree_depth(object)
                if my_depth + paste_depth > ConstructionFilter.maxdepth:
                    raise ValueError("Pasting would create a too deep structure.")


@indexer(IModule)
def SearchableTextIndexer(obj):
    """Index title, description and solution_direction."""
    return " ".join(
        [obj.title, StripMarkup(obj.description), StripMarkup(obj.solution_direction)]
    )


def tree_depth(obj):
    """Determine how deeply nested a module structure is.

    This is the opposite of item_depth.
    """
    submodules = [v for v in obj.values() if IQuestionContainer.providedBy(v)]
    if not submodules:
        return 1
    else:
        return 1 + max(tree_depth(m) for m in submodules)


def item_depth(item):
    """Return the survey depth of an item."""
    from euphorie.content.survey import ISurvey

    depth = 0
    for position in aq_chain(item):
        if ISurvey.providedBy(position):
            break
        depth += 1
    else:
        return None  # Not in a survey
    return depth


@adapter(ConditionalDexterityFTI, Interface)
@implementer(IConstructionFilter)
class ConstructionFilter:
    """FTI construction filter for :py:class:`Module` objects. This filter does
    two things: it restricts the maximum depth at which a module can be
    created, and it prevents creating of modules if the current container
    already contains a risk.

    This multi adapter requires the use of the conditional FTI as implemented
    by :py:class:`euphorie.content.fti.ConditionalDexterityFTI`.
    """

    maxdepth = 3

    def __init__(self, fti, container):
        self.fti = fti
        self.container = container

    def checkDepth(self):
        """Check if creating a new module would create a too deeply nested
        structure.

        :rtype: bool
        """
        depth = item_depth(self.container)
        if depth is None:
            return True
        return (depth + 1) <= self.maxdepth

    def checkForRisks(self):
        """Check if the container already contains a risk. If so refuse to
        allow creation of a module.

        :rtype: bool
        """
        for key in self.container:
            pt = self.container[key].portal_type
            if pt == "euphorie.risk":
                return False
        else:
            return True

    def allowed(self):
        """A module is allowed to be created providing :obj:`checkDepth` and
        :obj:`checkForRisks` are True

        :rtype: bool
        """
        return self.checkDepth() and self.checkForRisks()



from .. import MessageFactory as _
from plone.autoform.directives import order_after
from plone.autoform.interfaces import IFormFieldProvider
from plone.supermodel import model
from zope import schema
from zope.interface import provider


@provider(IFormFieldProvider)
class IHideFromTraining(model.Schema):
    hide_from_training = schema.Bool(
        title=_("label_hide_from_training", default="Hide from training"),
        description=_(
            "help_hide_from_training",
            default="Do not include this module in the training, for example "
            "if the module does not have any relevance for workers.",
        ),
        required=False,
        default=False,
    )
    order_after(hide_from_training="question")



from euphorie.client import model
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from plone.app.testing.interfaces import SITE_OWNER_NAME
from z3c.saconfig import Session


class RiskIntegrationTests(EuphorieIntegrationTestCase):
    def create_session_risk(self):
        with api.env.adopt_user(SITE_OWNER_NAME):
            api.content.create(
                container=self.portal.sectors, type="euphorie.country", id="eu"
            )
            client_country = api.content.create(
                container=self.portal.client, type="euphorie.clientcountry", id="eu"
            )
            client_sector = api.content.create(
                container=client_country, type="euphorie.clientsector", id="sector"
            )
            client_survey = api.content.create(
                container=client_sector, type="euphorie.survey", id="survey"
            )

        sqlsession = Session()
        account = model.Account(loginname="jane", password="secret")
        sqlsession.add(account)
        session = model.SurveySession(
            title="Session", zodb_path="eu/sector/survey", account=account
        )
        sqlsession.add(session)
        sqlsession.flush()

        module = session.addChild(
            model.Module(
                title="Root", module_id="1", zodb_path="1", has_description=False
            )
        )
        module = module.__of__(client_survey)
        risk = session.addChild(
            model.Risk(
                title="Risk 1",
                risk_id="1",
                zodb_path="1/1",
                type="risk",
                identification="no",
            )
        )
        risk = risk.__of__(module)

        return risk

    def test_default_collapsible_sections(self):
        risk = self.create_session_risk()

        with self._get_view("identification", risk) as view:
            # default default sections
            self.assertEqual(
                view.default_collapsible_sections, ["collapsible_section_information"]
            )

        # customized default sections
        view.webhelpers.content_country_obj.risk_default_collapsible_sections = [
            "collapsible_section_resources",
            "collapsible_section_comments",
        ]
        with self._get_view("identification", risk) as view:
            self.assertEqual(
                view.default_collapsible_sections,
                ["collapsible_section_resources", "collapsible_section_comments"],
            )

    def test_get_collapsible_section_state(self):
        risk = self.create_session_risk()

        with self._get_view("identification", risk) as view:
            # default default sections
            self.assertEqual(view.get_collapsible_section_state("information"), "")
            self.assertEqual(view.get_collapsible_section_state("resources"), "closed")
            self.assertEqual(view.get_collapsible_section_state("comments"), "closed")

        # customized default sections
        view.webhelpers.content_country_obj.risk_default_collapsible_sections = [
            "collapsible_section_resources",
            "collapsible_section_comments",
        ]
        with self._get_view("identification", risk) as view:
            self.assertEqual(
                view.get_collapsible_section_state("information"), "closed"
            )
            self.assertEqual(view.get_collapsible_section_state("resources"), "")
            self.assertEqual(view.get_collapsible_section_state("comments"), "")



"""Add the last_login column to the account table.

Revision ID: 20211019165915
Revises: 20211011114527
Create Date: 2021-10-19 16:59:15.046779
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20211019165915"
down_revision = "20211011114527"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("account", "last_login"):
        op.add_column("account", sa.Column("last_login", sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column("account", "last_login")



"""The boolean active column becomes the deactivated date time column.

Revision ID: 20211011114527
Revises: 20210409113814
Create Date: 2021-10-11 09:45:27.876273
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20211011114527"
down_revision = "20210409113814"
branch_labels = None
depends_on = None


def upgrade():
    """The boolean active column becomes the deactivated date time column."""
    if not has_column("group", "deactivated"):
        op.add_column("group", sa.Column("deactivated", sa.DateTime(), nullable=True))
        op.execute(
            """UPDATE "group" SET deactivated = '1970-01-01' WHERE active = FALSE"""
        )
    if has_column("group", "active"):
        op.drop_column("group", "active")


def downgrade():
    op.add_column(
        "group", sa.Column("active", sa.BOOLEAN(), autoincrement=False, nullable=True)
    )
    op.execute("""UPDATE "group" SET active = TRUE WHERE deactivated IS NULL""")
    op.execute("""UPDATE "group" SET active = FALSE WHERE deactivated IS NOT NULL""")
    op.drop_column("group", "deactivated")



"""
Survey views
------------
"""

from Acquisition import aq_chain
from Acquisition import aq_inner
from euphorie.client import model
from euphorie.client import utils
from euphorie.client.country import IClientCountry
from euphorie.client.profile import extractProfile
from euphorie.content.interfaces import ICustomRisksModule
from plone import api
from plone.memoize.instance import memoize
from plone.memoize.view import memoize_contextless
from sqlalchemy import case
from sqlalchemy import func
from sqlalchemy import orm
from sqlalchemy import sql
from z3c.saconfig import Session

import decimal
import logging


log = logging.getLogger(__name__)


class _StatusHelper:
    COUNTRIES_WITHOUT_HIGH_RISKS = ["it"]

    @property
    @memoize
    def sql_session(self):
        return Session()

    @property
    @memoize_contextless
    def preferred_language(self):
        return api.portal.get_tool("portal_languages").getPreferredLanguage()

    def module_query(self, sessionid, optional_modules):
        if optional_modules:
            case_clause = case(
                [
                    (
                        sql.and_(
                            model.SurveyTreeItem.profile_index != -1,
                            model.SurveyTreeItem.zodb_path.in_(optional_modules),
                        ),
                        func.substr(model.SurveyTreeItem.path, 1, 6),
                    ),
                    (
                        sql.and_(
                            model.SurveyTreeItem.profile_index == -1,
                            model.SurveyTreeItem.zodb_path.in_(optional_modules),
                        ),
                        func.substr(model.SurveyTreeItem.path, 1, 3) + "000-profile",
                    ),
                    (
                        sql.and_(
                            model.SurveyTreeItem.profile_index != -1,
                            model.SurveyTreeItem.depth < 2,
                        ),
                        func.substr(model.SurveyTreeItem.path, 1, 3),
                    ),
                ]
            )
        else:
            case_clause = case(
                [
                    (
                        sql.and_(
                            model.SurveyTreeItem.profile_index != -1,
                            model.SurveyTreeItem.depth < 2,
                        ),
                        func.substr(model.SurveyTreeItem.path, 1, 3),
                    )
                ]
            )

        query = (
            self.sql_session.query(case_clause.label("module"))
            .filter(
                sql.and_(
                    model.SurveyTreeItem.session_id == sessionid,
                    model.SurveyTreeItem.type == "module",
                )
            )
            .group_by("module")
            .order_by("module")
        )
        return query

    def slicePath(self, path):
        while path:
            yield path[:3].lstrip("0")
            path = path[3:]

    def getModulePaths(self):
        """Return a list of all the top-level modules belonging to this
        survey."""
        session_id = self.session.id
        if not session_id:
            return []
        profile = extractProfile(self.context.aq_parent, self.context.session)
        module_query = self.module_query(
            sessionid=session_id, optional_modules=profile.keys()
        )
        module_res = module_query.all()
        modules_and_profiles = {}
        for row in module_res:
            if row.module is not None:
                if row.module.find("profile") > 0:
                    path = row.module[:3]
                    modules_and_profiles[path] = "profile"
                else:
                    modules_and_profiles[row[0]] = ""
        module_paths = [m.module for m in module_res if m.module is not None]
        module_paths = modules_and_profiles.keys()
        module_paths = sorted(module_paths)
        self.modules_and_profiles = modules_and_profiles
        return module_paths

    def getModules(self):
        """Return a list of dicts of all the top-level modules and locations
        belonging to this survey."""
        sql_session = self.sql_session
        session_id = self.session.id
        module_paths = self.getModulePaths()
        url_schema = "%s/{0}/@@identification" % self.context.absolute_url()
        parent_node = orm.aliased(model.Module)
        titles = dict(
            sql_session.query(model.Module.path, model.Module.title)
            .filter(model.Module.session_id == session_id)
            .filter(model.Module.path.in_(module_paths))
        )

        location_titles = dict(
            sql_session.query(model.Module.path, parent_node.title)
            .filter(model.Module.session_id == session_id)
            .filter(model.Module.path.in_(module_paths))
            .filter(
                sql.and_(
                    parent_node.session_id == session_id,
                    parent_node.depth < model.Module.depth,
                    model.Module.path.like(parent_node.path + "%"),
                )
            )
        )
        modules = {}
        toc = {}
        title_custom_risks = utils.get_translated_custom_risks_title(self.request)

        for path in module_paths:
            number = ".".join(self.slicePath(path))
            # top-level module, always include it in the toc
            if len(path) == 3:
                title = titles[path]
                if (
                    title == "title_other_risks"
                    or title == "Other risks"
                    or title == "Custom risks"
                    or title == "label_custom_risks"
                ):
                    title = title_custom_risks
                toc[path] = {
                    "path": path,
                    "title": title,
                    "locations": [],
                    "number": number,
                }
                # If this is a profile (aka container for locations), skip
                # adding to the list of modules
                if self.modules_and_profiles[path] == "profile":
                    continue
            # sub-module (location) or location container
            else:
                if path in location_titles:
                    title = f"{location_titles[path]} - {titles[path]}"
                    toc[path[:3]]["locations"].append(
                        {"path": path, "title": titles[path], "number": number}
                    )
                else:
                    log.warning(
                        "Status: found a path for a submodule {} for which "
                        "there's no location title.".format(path)
                    )
                    continue

            modules[path] = {
                "path": path,
                "title": title,
                "url": url_schema.format("/".join(self.slicePath(path))),
                "todo": 0,
                "ok": 0,
                "postponed": 0,
                "risk_with_measures": 0,
                "risk_without_measures": 0,
                "number": number,
            }
        self.tocdata = toc
        return modules

    def getRisks(self, module_paths, skip_unanswered=False):
        """Return a list of risk dicts for risks that belong to the modules
        with paths as specified in module_paths."""
        global request
        request = self.request
        if not len(module_paths):
            return []
        sql_session = self.sql_session
        session_id = self.session.id
        # First, we need to compute the actual module paths, making sure that
        # skipped optional modules are excluded
        # This means top-level module paths like 001 or 001002 can be replaced
        # by several sub-modules paths like 001002, 001003 and 001002001
        path_clause = [model.SurveyTreeItem.path.like(f"{mp}%") for mp in module_paths]

        module_query = (
            sql_session.query(model.SurveyTreeItem)
            .filter(
                sql.and_(
                    model.SurveyTreeItem.session_id == session_id,
                    model.SurveyTreeItem.type == "module",
                    sql.or_(*path_clause),
                )
            )
            .order_by(model.SurveyTreeItem.path)
        )

        module_res = module_query.all()
        modules_by_path = {m.path: m for m in module_res}

        def nodes(modules):
            global use_nodes, s_paths
            use_nodes = []
            s_paths = sorted(modules, key=lambda x: x.path)
            # In case of repeatable profile questions, the top-level module
            # path will be 6 digits long.
            top_nodes = [
                elem
                for elem in s_paths
                if (
                    len(elem.path) == 3
                    or (len(elem.path) == 6 and elem.path[:3] not in s_paths)
                )
                and not elem.skip_children
            ]

            def use_node(elem):
                # Recursively find the nodes that are not disabled
                global use_nodes  # noqa: F824
                # Skip this elem?
                # If this is an optional module, check the "postponed" flag.
                # As long as the optional question has not been answered, skip
                # showing its children.
                # Only a "Yes" answer on the module will be considered as "do
                # not skip children"
                zodb_elem = self.context.aq_parent.restrictedTraverse(
                    elem.zodb_path.split("/"), None
                )
                if not zodb_elem:
                    log.error(
                        "Cannot traverse to %r from %r",
                        elem.zodb_path,
                        self.context.aq_parent,
                    )
                    return
                if getattr(zodb_elem, "optional", False):
                    if (
                        elem.postponed in (True, None) or elem.skip_children
                    ) and not ICustomRisksModule.providedBy(zodb_elem):
                        return
                children = [
                    x
                    for x in s_paths
                    if x.path.startswith(elem.path)
                    and len(x.path) == len(elem.path) + 3
                ]
                if children:
                    for child in children:
                        use_node(child)
                else:
                    use_nodes.append(elem.path)

            for elem in top_nodes:
                use_node(elem)
            ret = []
            # Here we make sure that only the longest paths of sub-modules
            # are used, but not the parents. Example
            # (001, 002, 001001, 001003) will be turned into
            # (001001, 001003, 002), since the parent 001 contains sub-modules,
            # and some of those might have been de-selected, like 001002
            for elem in sorted(use_nodes, reverse=True):
                if not [x for x in ret if x.startswith(elem)]:
                    ret.append(elem)
            return ret

        filtered_module_paths = nodes(tuple(module_res))

        child_node = orm.aliased(model.Risk)
        risks = (
            sql_session.query(model.Module, model.Risk)
            .filter(
                sql.and_(
                    model.Module.session_id == session_id,
                    model.Module.path.in_(filtered_module_paths),
                    sql.and_(
                        child_node.session_id == model.Module.session_id,
                        child_node.depth > model.Module.depth,
                        child_node.path.like(model.Module.path + "%"),
                    ),
                )
            )
            .join(
                (
                    model.Risk,
                    sql.and_(
                        model.Risk.path.startswith(model.Module.path),
                        model.Risk.session_id == session_id,
                    ),
                )
            )
            .order_by(model.Risk.path)
        )

        def _module_path(path):
            # Due to the extended query above that replaces top-module paths
            # with sub-module paths (if present), we need to cut back the path
            # under which we store each risk back to the original top-level
            # module path
            for mp in module_paths:
                if path.startswith(mp):
                    return mp

        filtered_risks = []
        for module, risk in risks.all():
            if risk.identification != "n/a":
                if (
                    skip_unanswered
                    and risk.identification is None
                    and not risk.postponed
                ):
                    continue
                module_path = _module_path(module.path)
                # And, since we might have truncated the path to represent
                # the top-level module, we also need to get the corresponding
                # module object.
                module = modules_by_path[module_path]
                filtered_risks.append((module, risk))
        return filtered_risks

    def as_decimal(self, num):
        return decimal.Decimal(num)

    @property
    @memoize
    def show_high_risks(self):
        for obj in aq_chain(aq_inner(self.context)):
            if IClientCountry.providedBy(obj):
                if obj.id in self.COUNTRIES_WITHOUT_HIGH_RISKS:
                    return False
        return True



from euphorie import MessageFactory as _
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import IFormLayer
from z3c.form.interfaces import IObjectWidget
from z3c.form.object import ObjectWidget
from z3c.form.widget import FieldWidget
from zope import schema
from zope.component import adapter
from zope.interface import classImplementsFirst
from zope.interface import implementer
from zope.interface import Interface
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class ISurveySourceSchema(Interface):
    source = schema.Choice(
        title=_("label_source", default="How would you like to start"),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm(
                    "scratch",
                    title=_(
                        "survey_source_scratch",
                        default="Create a new OiRA Tool from scratch",
                    ),
                ),
                SimpleTerm(
                    "local",
                    title=_(
                        "survey_source_local",
                        default=(
                            "Base my new OiRA Tool "
                            "on an existing OiRA Tool of my organisation"
                        ),
                    ),
                ),
                SimpleTerm(
                    "other",
                    title=_(
                        "survey_source_other",
                        default=(
                            "Base my new OiRA Tool "
                            "on an existing OiRA Tool of another organisation"
                        ),
                    ),
                ),
            ]
        ),
        default="scratch",
    )

    country = schema.Choice(
        title=_("label_choose_country", default="Choose a country"),
        vocabulary=SimpleVocabulary([]),
        required=False,
    )

    local_surveygroup = schema.Choice(
        title=_("label_choose_surveygroup", default="Choose an OiRA Tool"),
        required=False,
        vocabulary=SimpleVocabulary([]),
    )

    local_revision = schema.Choice(
        title=_(
            "label_choose_survey",
            default="Choose a revision for the selected OiRA Tool",
        ),
        required=False,
        vocabulary=SimpleVocabulary([]),
    )


class ISurveySourceSelectionField(schema.interfaces.IObject):
    """Interface for field that allows the user to select a source for the survey."""


class SurveySourceSelectionField(schema.Object):
    """A field that allows the user to select a source for the survey."""

    def __init__(self, **kwargs):
        super().__init__(schema=ISurveySourceSchema, **kwargs)


classImplementsFirst(SurveySourceSelectionField, ISurveySourceSelectionField)


class ISurveySourceSelectionWidget(IObjectWidget):
    """A widget for the survey source selection field."""


@implementer(ISurveySourceSelectionWidget)
class SurveySourceSelectionWidget(ObjectWidget):
    klass = "survey-source-selection-widget"
    css = "survey-source-selection"


@adapter(ISurveySourceSelectionField, IFormLayer)
@implementer(IFieldWidget)
def SurveySourceSelectionFieldWidget(field, request):
    """IFieldWidget factory for IObjectWidget."""
    return FieldWidget(field, SurveySourceSelectionWidget(request))



"""
Report
------

The landing page for the reports, plus the logic for creating the Excel report
for the action plan.
"""

from .. import MessageFactory as _
from datetime import date
from euphorie.client import model
from euphorie.client import survey
from euphorie.client import utils
from openpyxl.workbook import Workbook
from openpyxl.writer.excel import save_workbook
from plone import api
from plone.memoize.view import memoize
from Products.Five import BrowserView
from sqlalchemy import sql
from tempfile import NamedTemporaryFile
from urllib.parse import quote
from zope.i18n import translate

import logging


log = logging.getLogger(__name__)


class ReportLanding(BrowserView):
    """Custom report landing page.

    This replaces the standard online view of the report with a page
    offering the RTF and XLSX download options.
    """

    variation_class = "variation-risk-assessment"

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def default_reports(self):
        settings = self.webhelpers.content_country_obj
        default_reports = getattr(
            settings,
            "default_reports",
            ["report_full", "report_action_plan", "report_overview_risks"],
        )
        return default_reports

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        return self.index()


class ActionPlanTimeline(BrowserView, survey._StatusHelper):
    """Generate an excel file listing all measures.

    View name: @@timeline
    """

    @property
    def session(self):
        return self.context.session

    def get_measures(self):
        """Find all data that should be included in the report. Re-use the
        helper methods from Status to compute the filtered module paths and the
        relevant risks.

        The data is returned as a list of tuples containing a
        :py:class:`Module <euphorie.client.model.Module>`,
        :py:class:`Risk <euphorie.client.model.Risk>` and
        :py:class:`ActionPlan <euphorie.client.model.ActionPlan>`. Each
        entry in the list will correspond to a row in the generated Excel
        file.
        """
        module_paths = self.getModulePaths()
        risk_data = self.getRisks(module_paths)
        measure_data = []
        for module, risk in risk_data:
            action_plan_q = self.sql_session.query(model.ActionPlan).filter(
                sql.and_(
                    model.ActionPlan.risk_id == risk.id,
                    sql.or_(
                        model.ActionPlan.plan_type == "measure_standard",
                        model.ActionPlan.plan_type == "measure_custom",
                    ),
                )
            )
            # If the risk contains no action plan, add it as a single line
            # to the results
            action_plans = action_plan_q.all() or [None]
            for action_plan in action_plans:
                measure_data.append((module, risk, action_plan))

        # sort by 1. planning start, 2. path
        by_path_measure_data = sorted(measure_data, key=lambda x: x[1].path)
        return sorted(
            by_path_measure_data,
            key=lambda x: getattr(x[2], "planning_start", date.min) or date.min,
        )

    priority_names = {
        "low": _("label_timeline_priority_low", default="Low"),
        "medium": _("label_timeline_priority_medium", default="Medium"),
        "high": _("label_timeline_priority_high", default="High"),
    }

    columns = [
        (
            "measure",
            "planning_start",
            _("label_action_plan_start", default="Planning start"),
        ),
        (
            "measure",
            "planning_end",
            _("label_action_plan_end", default="Planning end"),
        ),
        (
            "measure",
            "action",
            _(
                "label_measure_action_plan",
                default="General approach " "(to eliminate or reduce the risk)",
            ),
        ),
        (
            "measure",
            "requirements",
            _(
                "label_measure_requirements",
                default="Level of expertise and/or requirements needed",
            ),
        ),
        (
            "measure",
            "responsible",
            _("label_action_plan_responsible", default="Who is responsible?"),
        ),
        ("measure", "budget", _("label_action_plan_budget", default="Budget")),
        ("module", "title", _("label_section", default="Section")),
        ("risk", "number", _("label_risk_number", default="Risk number")),
        ("risk", "title", _("report_timeline_risk_title", default="Risk")),
        ("risk", "priority", _("report_timeline_priority", default="Priority")),
        ("risk", "comment", _("report_timeline_comment", default="Comments")),
    ]

    def priority_name(self, priority):
        title = self.priority_names.get(priority)
        if title is not None:
            return translate(title, context=self.request)
        return priority

    def create_workbook(self):
        """Create an Excel workbook containing the all risks and measures."""
        book = Workbook()
        sheet = book.worksheets[0]
        sheet.title = translate(
            _("report_timeline_title", default="Timeline"), context=self.request
        )
        survey = self.context.aq_parent

        for column, (ntype, key, title) in enumerate(self.columns):
            sheet.cell(row=1, column=column + 1).value = translate(
                title, context=self.request
            )

        portal_transforms = api.portal.get_tool("portal_transforms")
        row = 2
        for module, risk, measure in self.get_measures():
            if risk.identification in ["n/a", "yes"]:
                continue

            column = 1
            if risk.is_custom_risk:
                zodb_node = None
            else:
                zodb_node = survey.restrictedTraverse(risk.zodb_path.split("/"))
            for ntype, key, title in self.columns:
                value = None
                if ntype == "measure":
                    value = getattr(measure, key, None)
                    if key == "action" and value:
                        value = portal_transforms.convertToData("text/plain", value)
                        if value:
                            value = value.strip()
                elif ntype == "risk":
                    value = getattr(risk, key, None)
                    if key == "comment":
                        value = portal_transforms.convertToData("text/plain", value)
                        if value:
                            value = value.strip()
                    elif key == "priority":
                        value = self.priority_name(value)
                    elif key == "title":
                        if (
                            getattr(zodb_node, "problem_description", None)
                            and zodb_node.problem_description.strip()
                        ):
                            value = zodb_node.problem_description
                    elif key == "number":
                        if risk.is_custom_risk:
                            num_elems = value.split(".")
                            value = ".".join(["Ω"] + num_elems[1:])
                elif ntype == "module":
                    if key == "title":
                        if risk.is_custom_risk:
                            value = utils.get_translated_custom_risks_title(
                                self.request
                            )
                        else:
                            value = module.title
                if value is not None:
                    cell = sheet.cell(row=row, column=column)
                    cell.value = value
                column += 1
            row += 1
        return book

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        book = self.create_workbook()
        filename = _(
            "filename_report_timeline",
            default="Action plan for ${title}",
            mapping={"title": self.session.title},
        )
        filename = translate(filename, context=self.request)
        self.request.response.setHeader(
            "Content-Disposition",
            "attachment; filename*=UTF-8''{}.xlsx".format(
                quote(filename.encode("utf-8"))
            ),
        )
        self.request.response.setHeader(
            "Content-Type",
            "application/vnd.openxmlformats-" "officedocument.spreadsheetml.sheet",
        )

        with NamedTemporaryFile() as tmp:
            save_workbook(book, tmp.name)
            tmp.seek(0)
            return tmp.read()



from euphorie.client.adapters.history_item import IHistoryItem
from euphorie.client.model import SessionEvent
from logging import getLogger
from plone.memoize.view import memoize
from Products.Five import BrowserView
from z3c.saconfig import Session
from zope.component import getAdapter
from zope.interface.interfaces import ComponentLookupError


logger = getLogger(__name__)


class HistoryPopup(BrowserView):
    """The view that keeps track of the history timeline of a session."""

    @property
    @memoize
    def events(self):
        """Get the events for this session."""
        return (
            Session.query(SessionEvent)
            .filter(SessionEvent.session == self.context.session)
            .order_by(SessionEvent.time)
        ).all()

    @property
    def items(self):
        """The sorted list of items related to this session."""
        session = self.context.session
        items = [getAdapter(session, IHistoryItem, name="started")]
        for event in self.events:
            try:
                items.append(getAdapter(event, IHistoryItem, name=event.action))
            except ComponentLookupError:
                logger.warning(
                    "Discarding unknown even type: %r, %r", event, event.action
                )
        if session.is_archived:
            items.append(getAdapter(session, IHistoryItem, name="archived"))
        return items



from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.content import MessageFactory as _
from Products.PlonePAS.plugins import passwordpolicy
from Products.PluggableAuthService.interfaces.plugins import IValidationPlugin
from zope import globalrequest
from zope.interface import implementer

import re


@implementer(IValidationPlugin)
class EuphoriePasswordPolicy(passwordpolicy.PasswordPolicyPlugin):
    """Simple Password Policy to ensure password is 5 chars long."""

    id = "euphorie_password_policy"
    meta_type = "Euphorie Password Policy"
    security = ClassSecurityInfo()

    security.declarePrivate("validateUserInfo")

    def validateUserInfo(self, user, set_id, set_info):
        """See IValidationPlugin.

        Used to validate password property
        """
        if IClientSkinLayer.providedBy(globalrequest.getRequest()):
            # We don't enforce the custom password policy for client users
            return super().validateUserInfo(user, set_id, set_info)

        if not set_info:
            return []
        password = set_info.get("password", None)
        if password is None:
            return []

        failed = False
        if len(password) < 5:
            failed = True
        elif len([letter for letter in password if letter.isupper()]) == 0:
            # Must have capital letter(s)
            failed = True
        elif not re.search("[1-9]", password):
            # Must have numbers(s)
            failed = True
        elif not re.search("[^a-zA-Z1-9]", password):
            # Must have special chars (i.e. not alphanumerical)
            failed = True

        if failed:
            return [
                {
                    "id": "password",
                    "error": _(
                        "password_policy_conditions",
                        default="Your password must contain at least 5 characters, "
                        "including at least one capital letter, one number and "
                        "one special character (e.g. $, # or @).",
                    ),
                }
            ]
        else:
            return []


InitializeClass(EuphoriePasswordPolicy)



"""
Login
-----

Register new users, login/logout, create a "Guest user" account and convert
existing guest accounts to normal accounts.
"""

from ..country import IClientCountry
from ..utils import setLanguage
from .conditions import approvedTermsAndConditions
from Acquisition import aq_chain
from Acquisition import aq_inner
from euphorie.client import config
from euphorie.client import MessageFactory as _
from euphorie.client import model
from euphorie.client.browser.country import SessionsView
from euphorie.client.model import get_current_account
from euphorie.content.survey import ISurvey
from plone import api
from plone.memoize.view import memoize
from plone.session.plugins.session import cookie_expiration_date
from plonetheme.nuplone.tiles.analytics import trigger_extra_pageview
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.PlonePAS.events import UserLoggedInEvent
from Products.statusmessages.interfaces import IStatusMessage
from urllib.parse import parse_qs
from urllib.parse import urlencode
from urllib.parse import urlparse
from urllib.parse import urlsplit
from z3c.saconfig import Session
from zExceptions import Unauthorized
from zope.lifecycleevent import notify

import datetime
import logging
import os
import re


log = logging.getLogger(__name__)

# I know this is a stupid regular expression, but it Works For Us(tm)
EMAIL_RE = re.compile(
    r"[_a-z0-9+-]+(\.[_a-z0-9+-]+)*@[a-z0-9-]+(\.[a-z0-9-]+)*(\.[a-z]{2,4})",
    re.IGNORECASE,
)


class Login(BrowserView):
    error = False
    errors = {}

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    def setLanguage(self, came_from):
        qs = urlparse(came_from)[4]
        params = parse_qs(qs)
        lang = params.get("language")
        if not lang:
            if IClientCountry.providedBy(self.context):
                lang = getattr(self.context, "language", None)
            elif ISurvey.providedBy(self.context):
                lang = self.context.language
        if lang:
            if isinstance(lang, str):
                lang = [lang]
            setLanguage(self.request, self.context, lang=lang[0])

    def login(self, account, remember):
        pas = getToolByName(self.context, "acl_users")
        with api.env.adopt_user(username=account.getUserName()):
            pas.updateCredentials(
                self.request,
                self.request.RESPONSE,
                account.loginname,
                account.password,
            )
        notify(UserLoggedInEvent(account))
        if remember:
            self.request.RESPONSE.cookies["__ac"]["expires"] = cookie_expiration_date(
                120
            )  # noqa: E501
            self.request.RESPONSE.cookies["__ac"]["max_age"] = (
                120 * 24 * 60 * 60
            )  # noqa: E501

    def transferGuestSession(self):
        """Transfer session(s) from guest account to an existing user account.

        The guest account is expected to go the login form as an authenticated user.

        For this reason we can know from the session plugin who he is even
        if during the request process we are now logged in as a new user.
        """
        new_account = get_current_account()
        if new_account is None:
            # We are not logged in with a postgres account,
            # so we cannot transfer any session
            return

        plugin = self.context.acl_users.session
        old_credentials = plugin.authenticateCredentials(
            plugin.extractCredentials(self.request)
        )
        if old_credentials is None:
            # We came here as anonymous user, no need to proceed
            return

        old_authenticated_account_id = old_credentials[0]

        try:
            old_authenticated_account_id = int(old_authenticated_account_id)
        except (ValueError, TypeError):
            # This happens when we login as a backend user
            # and the account id happens to be the email address
            # These users are never guests accounts, so we can skip the transfer
            return

        old_account = (
            Session.query(model.Account)
            .filter(
                model.Account.account_type == config.GUEST_ACCOUNT,
                model.Account.id == old_authenticated_account_id,
            )
            .first()
        )

        if old_account is None:
            # We check that the previously authenticated user was actually
            # a guest account that has been converted.
            # This prevents that a regular logged in user A logs in as B
            # and then all the sessions are transferred to B
            return

        sessions = Session.query(model.SurveySession).filter(
            model.SurveySession.account_id == old_authenticated_account_id
        )
        for session in sessions:
            session.account_id = new_account.id
            session.touch()
        Session.delete(old_account)

    def is_valid_password(self, password):
        if (
            len(password) < 12
            or not re.search("[A-Z]", password)
            or not re.search("[a-z]", password)
            or not re.search("[0-9]", password)
        ):
            return False
        return True

    def check_password_policy(self, password):
        if not self.is_valid_password(password):
            return _(
                "error_password_policy_violation",
                default=(
                    "The password needs to be at least 12 characters long and "
                    "needs to contain at least one lower case letter, one upper "
                    "case letter and one digit."
                ),
            )

    def _tryRegistration(self):
        if not self.webhelpers.allow_self_registration:
            raise Unauthorized("No self registration allowed.")
        form = self.request.form
        loginname = form.get("email")
        if not loginname:
            self.errors["email"] = _(
                "error_missing_email", default="Please enter your email address"
            )
        elif not EMAIL_RE.match(loginname):
            self.errors["email"] = _(
                "error_invalid_email", default="Please enter a valid email address"
            )
        if not form.get("password1"):
            self.errors["password"] = _(
                "error_missing_password", default="Please enter a password"
            )
        elif form.get("password1") != form.get("password2"):
            self.errors["password"] = _(
                "error_password_mismatch", default="Passwords do not match"
            )
        else:
            policy_error = self.check_password_policy(form.get("password1"))
            if policy_error:
                self.errors["password"] = policy_error
        if not form.get("terms"):
            self.errors["terms"] = _(
                "error_terms_not_accepted",
                default="An accout can only be created for you if you accept the "
                "terms and conditions.",
            )

        if self.errors:
            return False

        # Check honeypot fields
        if form.get("user_name") or form.get("user_email"):
            return False

        session = Session()
        loginname = loginname.lower()
        account = (
            session.query(model.Account)
            .filter(model.Account.loginname == loginname)
            .count()
        )
        if account:
            self.errors["email"] = _(
                "error_email_in_use",
                default="An account with this email address already exists.",
            )
            return False

        pm = getToolByName(self.context, "portal_membership")
        if pm.getMemberById(loginname) is not None:
            self.errors["email"] = _(
                "error_email_in_use",
                default="An account with this email address already exists.",
            )
            return False

        account = get_current_account()
        if account and account.account_type == config.GUEST_ACCOUNT:
            account.loginname = loginname
            account.password = form.get("password1")
            account.account_type = config.CONVERTED_ACCOUNT
            account.created = datetime.datetime.now()
            account.tc_approved = 1
            account.first_name = form.get("first_name")
            account.last_name = form.get("last_name")
            msg = _(
                "An account was created for you with email address ${email}",
                mapping={"email": loginname},
            )
            api.portal.show_message(msg, self.request, "success")
        else:
            account = model.Account(
                loginname=loginname,
                password=form.get("password1"),
                tc_approved=1,
                first_name=form.get("first_name"),
                last_name=form.get("last_name"),
            )
            Session().add(account)
        log.info("Registered new account %r", loginname)
        v_url = urlsplit(self.request.URL + "/success").path.replace("@@", "")
        trigger_extra_pageview(self.request, v_url)
        return account

    def get_or_create_account_for_backend_user(self):
        """Get or create an account for the currently logged in backend user.

        This is used when a backend user tries to log in to the client.

        The account will be created if there is no record in the database
        with the same email address as the backend user.

        The account will be returned when the record exists but the password used
        is different to the one stored in the database (e.g. is an ldap password).
        """
        form_email = self.request.form.get("__ac_name", "")
        if not form_email:
            return

        user = api.user.get_current()
        pas_email = user.getProperty("email", "")

        if form_email != pas_email:
            return

        account = (
            Session()
            .query(model.Account)
            .filter(model.Account.loginname == form_email)
            .first()
        )
        if account:
            return account

        password = self.request.form.get("__ac_password", "")
        if not password:
            reg = api.portal.get_tool(name="portal_registration")
            password = reg.generatePassword()

        account = model.Account(
            loginname=form_email,
            tc_approved=1,
            password=password,
            account_type=config.FULL_ACCOUNT,
        )
        Session().add(account)
        Session().flush()
        return account

    def __call__(self):
        context = aq_inner(self.context)
        self.errors = {}

        form = self.request.form

        came_from = self.webhelpers.get_came_from(default=self.webhelpers.country_url)
        self.setLanguage(came_from)

        account = get_current_account()

        self.allow_guest_accounts = api.portal.get_registry_record(
            "euphorie.allow_guest_accounts", default=False
        )
        lang = api.portal.get_current_language()

        self.show_whofor = False if lang in ("fr",) else True
        self.show_what_to_do = False if lang in ("fr",) else True
        self.show_how_long = False if lang in ("fr",) else True
        self.show_why_register = True
        self.show_prepare = False if lang in ("fr",) else True

        if self.request.method == "POST":
            if form.get("action") == "login":

                # Handle backend users that login to the client
                if not account and not api.user.is_anonymous():
                    account = self.get_or_create_account_for_backend_user()

                if (
                    isinstance(account, model.Account)
                    and account.getUserName() == form.get("__ac_name", "").lower()
                ):
                    self.transferGuestSession()
                    self.login(account, bool(self.request.form.get("remember")))
                    v_url = urlsplit(self.request.URL + "/success").path.replace(
                        "@@", ""
                    )
                    trigger_extra_pageview(self.request, v_url)
                else:
                    self.error = True

            elif form.get("action") == "register":
                account = self._tryRegistration()
                if account:
                    pas = getToolByName(self.context, "acl_users")
                    with api.env.adopt_user(username=account.getUserName()):
                        pas.updateCredentials(
                            self.request,
                            self.request.response,
                            account.getUserName(),
                            account.password,
                        )
                else:
                    self.error = True

            if not self.error:
                if api.portal.get_registry_record(
                    "euphorie.terms_and_conditions", default=False
                ) and not approvedTermsAndConditions(account):
                    self.request.RESPONSE.redirect(
                        "{}/terms-and-conditions?{}".format(
                            context.absolute_url(),
                            urlencode({"came_from": came_from}),
                        )
                    )
                else:
                    self.request.RESPONSE.redirect(came_from)
                return

        self.reset_password_request_url = "{}/@@reset_password_request?{}".format(
            context.absolute_url(),
            urlencode({"came_from": came_from}),
        )
        self.register_url = "{}/@@login#registration?{}".format(
            context.absolute_url(),
            urlencode({"came_from": came_from}),
        )
        self.tryout_url = "{}/@@tryout?{}".format(
            context.absolute_url(),
            urlencode({"came_from": came_from}),
        )

        return self.index()

    def get_image_version(self, name):
        """Needed on the reports overview shown to the guest user (view name:

        @@register_session)
        """
        fdir = os.path.join(
            os.path.dirname(__file__), os.path.join("..", "resources", "media")
        )
        lang = getattr(self.request, "LANGUAGE", "en")
        fname = f"{name}_{lang}"
        if os.path.isfile(os.path.join(fdir, fname + ".png")):
            return fname
        return name


class Tryout(SessionsView, Login):
    """Create a guest account.

    View name: @@tryout
    """

    def createGuestAccount(self):
        account = model.Account(
            loginname="guest-%s" % datetime.datetime.now().isoformat(),
            account_type=config.GUEST_ACCOUNT,
        )
        Session().add(account)
        return account

    def __call__(self):
        came_from = self.request.form.get("came_from")
        if not came_from:
            return self.request.response.redirect(api.portal.get().absolute_url())

        # Check honeypot fields
        if self.request.form.get("user_name") or self.request.form.get("user_email"):
            return self.request.response.redirect(api.portal.get().absolute_url())

        account = self.createGuestAccount()
        self.login(account, False)
        client_url = self.request.client.absolute_url()
        came_from = came_from.replace(client_url, "")
        if came_from.startswith("/"):
            came_from = came_from[1:]
        try:
            survey = self.context.restrictedTraverse(came_from)
        except KeyError:
            survey = None
        # This might happen if we're linking to a dedicated session. Try to
        # strip the session information from the came_from and try again
        except (AttributeError, Unauthorized):
            came_from = came_from.split("++session++")[0]
            try:
                survey = self.context.restrictedTraverse(came_from)
            except KeyError:
                survey = None
        if not ISurvey.providedBy(survey):
            return self.request.response.redirect(came_from)
        info = dict(action="new", survey=came_from)
        self._NewSurvey(info, account)


class CreateTestSession(Tryout):
    """Create a guest session.

    View name: @@new-session-test.html
    """

    def __call__(self):
        self.allow_guest_accounts = api.portal.get_registry_record(
            "euphorie.allow_guest_accounts", default=False
        )
        context = aq_inner(self.context)
        webhelpers = api.content.get_view(
            name="webhelpers", context=self.context, request=self.request
        )
        came_from = webhelpers.get_came_from(default=context.absolute_url())
        self.register_url = "{}/@@login?{}#registration".format(
            context.absolute_url(),
            urlencode({"came_from": came_from}),
        )
        setLanguage(self.request, self.context)
        if self.request.environ["REQUEST_METHOD"] == "POST":
            if not self.allow_guest_accounts:
                flash = IStatusMessage(self.request).addStatusMessage
                flash(
                    _(
                        "Starting a test session is not available in this OiRA "
                        "application."
                    ),
                    "error",
                )
                self.request.response.redirect(came_from)
            else:
                form = self.request.form
                if form["action"] == "new":
                    account = self.createGuestAccount()
                    self.login(account, False)
                    self._NewSurvey(form, account)
        self._updateSurveys()
        return self.index()


class Logout(BrowserView):
    def __call__(self):
        pas = getToolByName(self.context, "acl_users")
        pas.resetCredentials(self.request, self.request.response)

        for obj in aq_chain(aq_inner(self.context)):
            if IClientCountry.providedBy(obj):
                break
        else:
            obj = self.request.client

        self.request.response.redirect(obj.absolute_url())



"""Add the scaled_answer column to the risk table

Revision ID: 20240419142030
Revises: 20231019121221
Create Date: 2024-04-19 14:19:41.050854

"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20240419142030"
down_revision = "20231019121221"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    if not has_column("risk", "scaled_answer"):
        op.add_column(
            "risk", sa.Column("scaled_answer", sa.UnicodeText(), nullable=True)
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    if has_column("risk", "scaled_answer"):
        op.drop_column("risk", "scaled_answer")
    # ### end Alembic commands ###



from euphorie.client.tests.utils import addSurvey
from euphorie.client.tests.utils import registerUserInClient
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieFunctionalTestCase
from urllib.parse import urlencode


class CountryFunctionalTests(EuphorieFunctionalTestCase):
    def test_surveys_filtered_by_language(self):
        survey = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
                      <title>Sector</title>
                      <survey>
                        <title>Survey</title>
                        <language>en</language>
                      </survey>
                    </sector>"""
        survey_nl = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
                    <title>Branche</title>
                    <survey>
                      <title>Vragenlijst</title>
                      <language>nl</language>
                    </survey>
                  </sector>"""  # noqa
        self.loginAsPortalOwner()
        addSurvey(self.portal, survey)
        addSurvey(self.portal, survey_nl)
        browser = self.get_browser()
        # Pass the language as URL parameter to ensure that we get the NL
        # version
        browser.open("%s?language=nl" % self.portal.client.absolute_url())
        registerUserInClient(browser, link="Registreer")
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        self.assertEqual(
            browser.getControl(name="survey").options, ["branche/vragenlijst"]
        )

        # Still Dutch
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        self.assertEqual(
            browser.getControl(name="survey").options, ["branche/vragenlijst"]
        )

        # Now, switch to English
        browser.open("%s?language=en" % self.portal.client["nl"].absolute_url())
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        self.assertEqual(browser.getControl(name="survey").options, ["sector/survey"])

    def test_must_select_valid_survey(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        browser = self.get_browser()
        browser.open(self.portal.client["nl"].absolute_url())
        registerUserInClient(browser, link="Register")
        data = urlencode({"action": "new", "survey": "", "title:utf8:ustring": "Foo"})
        browser.handleErrors = False
        browser.open(browser.url, data)
        self.assertEqual(browser.url, "http://nohost/plone/client/nl")



from euphorie.htmllaundry.utils import strip_markup
from lxml import etree

import docx
import lxml.html


# tag to style mapping which will be used over styles from the source document.
ENFORCE_STYLES = {
    "ul": "List Bullet",
    "ol": "List Number",
}


def add_hyperlink(paragraph, url, text, style):
    """A function that places a hyperlink within a paragraph object.

    :param paragraph: The paragraph we are adding the hyperlink to.
    :param url: A string containing the required url
    :param text: The text displayed for the url
    :return: The hyperlink object
    """

    # This gets access to the document.xml.rels file and gets a new relation id value
    part = paragraph.part
    r_id = part.relate_to(
        url, docx.opc.constants.RELATIONSHIP_TYPE.HYPERLINK, is_external=True
    )

    # Create the w:hyperlink tag and add needed values
    hyperlink = docx.oxml.shared.OxmlElement("w:hyperlink")
    hyperlink.set(
        docx.oxml.shared.qn("r:id"),
        r_id,
    )

    # Create a w:r element
    new_run = docx.oxml.shared.OxmlElement("w:r")

    # Create a new w:rPr element
    rPr = docx.oxml.shared.OxmlElement("w:rPr")

    # Join all the xml elements together add add the required text to the w:r element
    new_run.append(rPr)
    new_run.text = text
    new_run.style = style
    hyperlink.append(new_run)

    paragraph._p.append(hyperlink)

    return hyperlink


class HtmlToWord:
    def handleInlineText(self, node, p):
        """Handler for elements which can only contain inline text (p, li)"""
        run = p.add_run()
        font = run.font
        if node.tag in ["strong", "b"]:
            font.bold = True
        elif node.tag in ["em", "i"]:
            font.italic = True
        elif node.tag == "u":
            font.underline = True

        if node.text and node.text.strip():
            if node.tag == "a":
                href = node.get("href")
                href = href and href.strip()
                if href and href != node.text.strip():
                    add_hyperlink(p, href, node.text, "Hyperlink")
                else:
                    run.text = node.text
            else:
                run.text = node.text

        for sub in node:
            p = self.handleInlineText(sub, p)
        if node.tail and node.tail.strip():
            run = p.add_run()
            run.text = node.tail
        return p

    def handleElement(self, node, doc, style=None):
        style = ENFORCE_STYLES.get(node.tag) or style

        if node.tag in [
            "a",
            "b",
            "blockquote",
            "em",
            "h4",
            "i",
            "li",
            "p",
            "strong",
            "u",
        ]:
            p = doc.add_paragraph(style=style)
            p = self.handleInlineText(node, p)
        elif node.tag in ["ul", "ol"]:
            for sub in node:
                if sub.tag == "li":
                    p = doc.add_paragraph(style=style)
                    p.paragraph_format.left_indent = docx.shared.Inches(1)
                    p = self.handleInlineText(sub, p)

        tail = node.tail
        # Prevent unwanted empty lines inside listings and paragraphs that come
        # from newlines in the markup
        # if node.tag in ['li', 'p', 'strong', 'em', 'b', 'i']:
        tail = tail and tail.strip()
        if tail and not doc.paragraphs[-1].text.endswith(tail):
            doc.add_paragraph(tail)
        return doc

    def __call__(self, markup, doc, style=None, next_style=None):
        if not markup or not markup.strip():
            return doc
        try:
            markup_doc = lxml.html.document_fromstring(markup)
        except etree.XMLSyntaxError:
            text = strip_markup(markup)
            text = text.replace("&#13", "\n")
            doc.add_paragraph(text)
            return doc

        for idx, node in enumerate(markup_doc.find("body")):
            if idx == 0:
                p_style = style
            else:
                p_style = next_style or style
            doc = self.handleElement(node, doc, p_style)

        return doc



"""Create the organisation_membership table.

Revision ID: 20220720110538
Revises: 20220705145832
Create Date: 2022-04-01 14:22:44.091422
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column
from euphorie.deployment.upgrade.utils import has_table

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20220720110538"
down_revision = "20220705145832"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("organisation_membership"):
        op.create_table(
            "organisation_membership",
            sa.Column(
                "organisation_id", sa.Integer(), autoincrement=True, nullable=False
            ),
            sa.Column("owner_id", sa.Integer(), nullable=False),
            sa.Column("member_id", sa.Integer(), nullable=False),
            sa.ForeignKeyConstraint(
                ["member_id"], ["account.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.ForeignKeyConstraint(
                ["owner_id"], ["account.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("organisation_id"),
        )

    if not has_column("organisation_membership", "member_role"):
        op.add_column(
            "organisation_membership",
            sa.Column("member_role", sa.UnicodeText(), nullable=True),
        )

    if not has_table("organisation"):
        op.create_table(
            "organisation",
            sa.Column(
                "organisation_id", sa.Integer(), autoincrement=True, nullable=False
            ),
            sa.Column("owner_id", sa.Integer(), nullable=False),
            sa.Column("title", sa.UnicodeText(), nullable=True),
            sa.Column("image_data", sa.LargeBinary(), nullable=True),
            sa.Column("image_data_scaled", sa.LargeBinary(), nullable=True),
            sa.Column("image_filename", sa.UnicodeText(), nullable=True),
            sa.ForeignKeyConstraint(
                ["owner_id"], ["account.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("organisation_id"),
        )


def downgrade():
    if has_table("organisation"):
        op.drop_table("organisation")
    if has_table("organisation_membership"):
        op.drop_table("organisation_membership")



import unittest


class Mock:
    pass


class strip_markup_tests(unittest.TestCase):
    def strip_markup(self, *a, **kw):
        from euphorie.htmllaundry.utils import strip_markup

        return strip_markup(*a, **kw)

    def testEmpty(self):
        obj = Mock()
        obj.description = ""
        self.assertEqual(self.strip_markup(""), "")

    def testNoMarkup(self):
        self.assertEqual(self.strip_markup("Test"), "Test")

    def testSingleTag(self):
        self.assertEqual(self.strip_markup("Test <em>me</me>"), "Test me")

    def testMultipleTags(self):
        self.assertEqual(
            self.strip_markup("Test <em>me</me> <strong>now</strong>"), "Test me now"
        )

    def testStrayBracket(self):
        self.assertEqual(self.strip_markup("Test <em>me</em> >"), "Test me >")


class remove_empty_tags_tests(unittest.TestCase):
    def _remove(self, str, extra_tags=[]):
        from euphorie.htmllaundry.utils import remove_empty_tags

        import lxml.etree

        fragment = lxml.etree.fromstring(str)
        fragment = remove_empty_tags(fragment, extra_tags)
        return lxml.etree.tostring(fragment, encoding="utf8").decode()

    def testRemoveEmptyParagraphElement(self):
        self.assertEqual(self._remove("<div><p/></div>"), "<div/>")

    def testRemoveEmptyParagraph(self):
        self.assertEqual(self._remove("<div><p></p></div>"), "<div/>")

    def testRemoveParagraphWithWhitespace(self):
        self.assertEqual(self._remove("<div><p>  </p></div>"), "<div/>")

    def testRemoveParagraphWithUnicodeWhitespace(self):
        self.assertEqual(self._remove("<div><p> \xa0 </p></div>"), "<div/>")

    def testKeepEmptyImageElement(self):
        self.assertEqual(
            self._remove('<div><img src="image"/></div>'),
            '<div><img src="image"/></div>',
        )

    def testCollapseBreaks(self):
        self.assertEqual(
            self._remove("<body><p>one<br/><br/>two</p></body>"),
            "<body><p>one<br/>two</p></body>",
        )

    def testNestedData(self):
        self.assertEqual(
            self._remove("<div><h3><bad/></h3><p>Test</p></div>"),
            "<div><p>Test</p></div>",
        )

    def testKeepElementsWithTail(self):
        self.assertEqual(
            self._remove("<body>One<br/> two<br/> three</body>"),
            "<body>One<br/> two<br/> three</body>",
        )

    def testTrailingBreak(self):
        self.assertEqual(self._remove("<div>Test <br/></div>"), "<div>Test </div>")

    def testLeadingBreak(self):
        self.assertEqual(self._remove("<div><br/>Test</div>"), "<div>Test</div>")

    def testDoNotRemoveEmptyAnchorElement(self):
        # Should not remove empty <a> tag because it's used as an anchor:
        self.assertEqual(
            self._remove('<p><a id="anchor"></a></p>'), '<p><a id="anchor"/></p>'
        )
        self.assertEqual(
            self._remove('<p><a name="anchor" /></p>'), '<p><a name="anchor"/></p>'
        )
        self.assertEqual(
            self._remove('<p><a href="blah" id="anchor"/></p>'),
            '<p><a href="blah" id="anchor"/></p>',
        )
        self.assertEqual(
            self._remove('<p><a href="blah" name="anchor"/></p>'),
            '<p><a href="blah" name="anchor"/></p>',
        )

        # Should not remove <a> tag because it's non-empty:
        self.assertEqual(
            self._remove('<p><a href="blah" name="anchor">Link</a></p>'),
            '<p><a href="blah" name="anchor">Link</a></p>',
        )
        self.assertEqual(
            self._remove('<p><a name="anchor">Link</a></p>'),
            '<p><a name="anchor">Link</a></p>',
        )
        self.assertEqual(
            self._remove('<p><a href="anchor">Link</a></p>'),
            '<p><a href="anchor">Link</a></p>',
        )

        # Should remove because it's an useless empty tag.
        self.assertEqual(
            self._remove('<p><a href="blah"/>Content</p>'), "<p>Content</p>"
        )
        self.assertEqual(
            self._remove('<p><a href="blah"></a>Content</p>'), "<p>Content</p>"
        )

    def testExtraAllowedEmptyTags(self):
        self.assertEqual(
            self._remove("<table><tr><td>Test</td><td></td></tr></table>", ["td"]),
            "<table><tr><td>Test</td><td/></tr></table>",
        )


class ForceLinkTargetTests(unittest.TestCase):
    def force_link_target(self, str, target="_blank"):
        from euphorie.htmllaundry.cleaners import LaundryCleaner

        import lxml.etree

        fragment = lxml.etree.fromstring(str)
        cleaner = LaundryCleaner()
        cleaner.force_link_target(fragment, target)
        return lxml.etree.tostring(fragment, encoding="utf8").decode()

    def testNoAnchor(self):
        self.assertEqual(self.force_link_target("<div><p/></div>"), "<div><p/></div>")

    def testAddTarget(self):
        self.assertEqual(
            self.force_link_target(
                '<div><a href="http://example.com"/></div>', "_blank"
            ),
            '<div><a href="http://example.com" target="_blank"/></div>',
        )

    def testRemoveTarget(self):
        self.assertEqual(
            self.force_link_target(
                '<div><a target="_blank" href="http://example.com"/></div>', None
            ),
            '<div><a href="http://example.com"/></div>',
        )


class strip_outer_breaks_tests(unittest.TestCase):
    def _strip(self, str):
        from euphorie.htmllaundry.utils import strip_outer_breaks

        import lxml.etree

        fragment = lxml.etree.fromstring(str)
        strip_outer_breaks(fragment)
        return lxml.etree.tostring(fragment, encoding="utf8").decode()

    def testNoBreak(self):
        self.assertEqual(
            self._strip("<body>Dummy text</body>"), "<body>Dummy text</body>"
        )

    def testTrailingBreak(self):
        self.assertEqual(
            self._strip("<body>Dummy text<br/></body>"), "<body>Dummy text</body>"
        )

    def testLeadingBreak(self):
        self.assertEqual(
            self._strip("<body><br/>Dummy text</body>"), "<body>Dummy text</body>"
        )

    def testBreakAfterElement(self):
        self.assertEqual(
            self._strip("<body><p>Dummy</p><br/>text</body>"),
            "<body><p>Dummy</p>text</body>",
        )


class SanizeTests(unittest.TestCase):
    def sanitize(self, *a, **kw):
        from euphorie.htmllaundry.utils import sanitize

        return sanitize(*a, **kw)

    def testEmpty(self):
        self.assertEqual(self.sanitize(""), "")

    def testParagraph(self):
        self.assertEqual(self.sanitize("<p>Test</p>"), "<p>Test</p>")

    def test_link_in_unwrapped_text(self):
        self.assertEqual(
            self.sanitize('There is a <a href="#">link</a> in here.'),
            '<p>There is a <a href="#">link</a> in here.</p>',
        )

    def testParagraphCustomWrapperNotUsedIfAlreadyWrapped(self):
        self.assertEqual(self.sanitize("<p>Test</p>", wrap="span"), "<p>Test</p>")

    def testParagraphWithWhitespace(self):
        self.assertEqual(self.sanitize("<p>Test</p>\n<p>\xa0</p>\n"), "<p>Test</p>\n\n")

    def testLeadingBreak(self):
        self.assertEqual(self.sanitize("<br/><p>Test</p>"), "<p>Test</p>")

    def testHeaderAndText(self):
        self.assertEqual(
            self.sanitize("<h3>Title</h3><p>Test</p>"), "<h3>Title</h3><p>Test</p>"
        )

    def testUnwrappedText(self):
        self.assertEqual(self.sanitize("Hello, World"), "<p>Hello, World</p>")

    def testUnwrappedTextWithCustomWrapper(self):
        self.assertEqual(
            self.sanitize("Hello, World", wrap="strong"),
            "<strong>Hello, World</strong>",
        )

    def testTrailingUnwrappedText(self):
        self.assertEqual(
            self.sanitize("<p>Hello,</p> World"), "<p>Hello,</p><p> World</p>"
        )

    def testTrailingUnwrappedTextWithCustomWrapper(self):
        self.assertEqual(
            self.sanitize("<p>Hello,</p> World", wrap="b"), "<p>Hello,</p><b> World</b>"
        )

    def testUnwrappedTextEverywhere(self):
        self.assertEqual(
            self.sanitize(
                "Hello, <p>you</p> nice and <strong>decent</strong> <em>person</em>."
            ),
            "<p>Hello, </p><p>you</p><p> nice and <strong>decent</strong> "
            + "<em>person</em>.</p>",
        )

    def testUnwrappedTextEverywhereWithCustomWrapper(self):
        self.assertEqual(
            self.sanitize("Hello, <p>you</p> nice <em>person</em>.", wrap="div"),
            "<div>Hello, </div><p>you</p><div> nice " + "<em>person</em>.</div>",
        )

    def testStripStyleAttributes(self):
        self.assertEqual(
            self.sanitize('<p style="text-weight: bold">Hello</p>'), "<p>Hello</p>"
        )

    def testJavascriptLink(self):
        self.assertEqual(
            self.sanitize(
                "<p><a href=\"javascript:alert('I am evil')\">" + "click me</a></p>"
            ),
            '<p><a href="">click me</a></p>',
        )

    def testSkipWrapping(self):
        self.assertEqual(
            self.sanitize("Hello, <em>you</em> nice <em>person</em>.", wrap=None),
            "Hello, <em>you</em> nice <em>person</em>.",
        )

    def testRejectBadWrapElement(self):
        self.assertRaises(ValueError, self.sanitize, "<p>Hello,</p> World", wrap="xxx")
        self.assertRaises(
            ValueError,
            self.sanitize,
            "Hello, <em>you</em> nice <em>person</em>.",
            wrap="",
        )



"""Add a timestamp to the company table.

Revision ID: 20210409113814
Revises: 29
Create Date: 2021-04-09 11:58:37.115736
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20210409113814"
down_revision = "29"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("company", "timestamp"):
        op.add_column("company", sa.Column("timestamp", sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column("company", "timestamp")



from ..fti import check_fti_paste_allowed
from euphorie.content.fti import ConditionalDexterityFTI
from euphorie.content.fti import IConstructionFilter
from euphorie.content.module import Module
from euphorie.testing import EuphorieIntegrationTestCase
from unittest import mock
from zope.component import adapter
from zope.component import getGlobalSiteManager
from zope.component import provideAdapter
from zope.interface import implementer

import unittest


@implementer(IConstructionFilter)
@adapter(ConditionalDexterityFTI, Module)
class Veto:
    def __init__(self, fti, container):
        self.fti = fti
        self.container = container

    def allowed(self):
        return False


class ConditionalDexterityFTITests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()

    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createContent(self):
        self.country = self.portal.sectors.nl
        self.sector = self._create(self.country, "euphorie.sector", "sector")
        self.surveygroup = self._create(self.sector, "euphorie.surveygroup", "group")
        self.survey = self._create(self.surveygroup, "euphorie.survey", "survey")
        self.module = self._create(self.survey, "euphorie.module", "survey")

    def testConditionalFtiUsedByModule(self):
        fti = getattr(self.portal.portal_types, "euphorie.module")
        self.assertTrue(isinstance(fti, ConditionalDexterityFTI))

    def testVeto(self):
        self.createContent()
        provideAdapter(Veto, name="euphorie.module")
        try:
            fti = getattr(self.portal.portal_types, "euphorie.module")
            self.assertEqual(fti.isConstructionAllowed(self.module), False)
        finally:
            getGlobalSiteManager().unregisterAdapter(Veto, name="euphorie.module")

    def testVetoRegisteredForOtherType(self):
        self.createContent()
        provideAdapter(factory=Veto, name="euphorie.solution")
        try:
            fti = getattr(self.portal.portal_types, "euphorie.module")
            self.assertEqual(fti.isConstructionAllowed(self.module), True)
        finally:
            getGlobalSiteManager().unregisterAdapter(Veto, name="euphorie.solution")


class check_fti_paste_allowed_tests(unittest.TestCase):
    def check_fti_paste_allowed(self, *a, **kw):
        return check_fti_paste_allowed(*a, **kw)

    def test_acceptable_paste(self):
        content = mock.Mock()
        content.portal_type = "euphorie.risk"
        fti = mock.Mock()
        fti.isConstructionAllowed.return_value = True
        with mock.patch("euphorie.content.fti.queryUtility", return_value=fti):
            self.check_fti_paste_allowed("folder", content)
            fti.isConstructionAllowed.assert_called_once_with("folder")

    def test_refuse_non_portal_content(self):
        self.assertRaises(ValueError, self.check_fti_paste_allowed, None, mock.Mock())

    def test_content_without_fti(self):
        content = mock.Mock()
        content.portal_type = "euphorie.risk"
        self.assertRaises(ValueError, self.check_fti_paste_allowed, None, content)

    def test_fti_does_not_allow_construction(self):
        content = mock.Mock()
        content.portal_type = "euphorie.risk"
        fti = mock.Mock()
        fti.isConstructionAllowed.return_value = False
        with mock.patch("euphorie.content.fti.queryUtility", return_value=fti):
            self.assertRaises(
                ValueError, self.check_fti_paste_allowed, "folder", content
            )



from Acquisition import aq_base
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.model import Account
from euphorie.testing import EuphorieIntegrationTestCase
from plonetheme.nuplone.skin.interfaces import NuPloneSkin
from z3c.saconfig import Session
from zope.interface import alsoProvides
from zope.interface import noLongerProvides


class AuthenticationTests(EuphorieIntegrationTestCase):
    def createAccount(self, login="john", password="jane"):
        session = Session()
        account = Account(loginname=login, password=password)
        session.add(account)
        session.flush()
        return account

    def testGetUserById_UnknownAccount(self):
        self.assertEqual(self.portal.acl_users.getUserById("john"), None)

    def testGetUserById_ValidAccount(self):
        request = self.app.REQUEST
        alsoProvides(request, IClientSkinLayer)
        account = self.createAccount()
        user = self.portal.acl_users.getUserById(str(account.id))
        self.assertTrue(aq_base(user) is account)
        self.assertTrue(isinstance(user.getId(), str))

    def testChallenge_OutsideClient(self):
        self.logout()
        request = self.request
        request._has_challenged = False
        # XXX the NuPloneSkin challenger returns a 403
        noLongerProvides(request, NuPloneSkin)
        self.portal.acl_users(None, request)
        self.portal.acl_users._unauthorized()
        self.assertEqual(
            request.response.headers["location"],
            "http://nohost/plone/acl_users/credentials_cookie_auth/require_login?came_from=",  # noqa: E501
        )

    def testChallenge_InClient(self):
        self.logout()
        request = self.app.REQUEST
        request["PUBLISHED"] = self.portal
        request._has_challenged = False
        alsoProvides(request, IClientSkinLayer)
        self.portal.acl_users(None, request)
        self.portal.acl_users._unauthorized()
        self.assertEqual(
            request.response.headers["location"],
            "http://nohost/plone/@@login?came_from=http%3A%2F%2Fnohost#login",
        )



"""
Help
----

A Help document with multiple rich text fields.

https://admin.oiraproject.eu/documents/en/help

portal_type: euphorie.help
"""

from .. import MessageFactory as _
from euphorie.content.utils import StripMarkup
from euphorie.htmllaundry.z3cform import HtmlText
from plone.autoform import directives
from plone.indexer import indexer
from plone.supermodel import model
from plonetheme.nuplone.z3cform.widget import WysiwygFieldWidget


class IOnlineHelp(model.Schema):
    introduction = HtmlText(
        title=_("label_help_introduction", default="Introduction"),
        description=_(
            "help_help_introduction", default="General information on risk assessment"
        ),
        required=True,
    )
    directives.widget(introduction=WysiwygFieldWidget)

    authentication = HtmlText(
        title=_("label_help_authentication", default="Registration"),
        description=_(
            "help_authentication",
            default="This text should explain how to register and login.",
        ),
        required=True,
    )
    directives.widget(authentication=WysiwygFieldWidget)

    sessions = HtmlText(
        title=_("label_help_sessions", default="Carrying out your risk assessment"),
        description=_(
            "help_sessions",
            default="This text should describe the main functions of the OiRA Tool.",
        ),
        required=True,
    )
    directives.widget(sessions=WysiwygFieldWidget)

    preparation = HtmlText(
        title=_("label_help_preparation", default="1. Preparation"),
        description=_(
            "help_preparation",
            default="This text should explain the 2 types of profile questions.",
        ),
        required=True,
    )
    directives.widget(preparation=WysiwygFieldWidget)

    identification = HtmlText(
        title=_("label_help_identification", default="2. Identification"),
        description=_(
            "help_identification",
            default="This text should explain how the risk identification works.",
        ),
        required=True,
    )
    directives.widget(identification=WysiwygFieldWidget)

    evaluation = HtmlText(
        title=_("label_help_evaluation", default="3. Evaluation"),
        description=_(
            "help_evaluation",
            default="This text should explain how to evaluate the identified risks.",
        ),
        required=True,
    )
    directives.widget(evaluation=WysiwygFieldWidget)

    actionplan = HtmlText(
        title=_("label_help_actionplan", default="4. Action Plan"),
        description=_(
            "help_actionplan",
            default="This text should explain how to fill in the Action plan.",
        ),
        required=True,
    )
    directives.widget(actionplan=WysiwygFieldWidget)

    report = HtmlText(
        title=_("label_help_reports", default="5. Report"),
        description=_(
            "help_reports",
            default="This text should describe how the report can either be saved or printed.",  # noqa: E501
        ),
        required=True,
    )
    directives.widget(report=WysiwygFieldWidget)

    finalwords = HtmlText(
        title=_("label_help_finalwords", default="What happens next?"),
        description=_("help_finalwords", default="General final recommendations."),
        required=True,
    )
    directives.widget(finalwords=WysiwygFieldWidget)


@indexer(IOnlineHelp)
def SearchableTextIndexer(obj):
    """Index the introduction, authentication, sessions, identification,
    evaluation, actionplan, report and finalwords."""
    return " ".join(
        [
            StripMarkup(obj.introduction),
            StripMarkup(obj.authentication),
            StripMarkup(obj.sessions),
            StripMarkup(obj.identification),
            StripMarkup(obj.evaluation),
            StripMarkup(obj.actionplan),
            StripMarkup(obj.report),
            StripMarkup(obj.finalwords),
        ]
    )



from Acquisition import aq_base
from euphorie.client.browser.webhelpers import WebHelpers
from euphorie.client.country import IClientCountry


class ClientView(WebHelpers):
    def __call__(self):
        """The frontpage has been disbanded.

        We redirect to the country that is defined as the default, or
        pick a random country.
        """
        target = None
        language = self.request.form.get("language")
        url_param = language and "?language=%s" % language or ""
        if self.default_country:
            if getattr(aq_base(self.context), self.default_country, None):
                found = getattr(self.context, self.default_country)
                if IClientCountry.providedBy(found):
                    target = found
        if not target:
            for id, found in self.context.objectItems():
                if IClientCountry.providedBy(found):
                    target = found
                    break
        if not target:
            return "No country was identified"
        self.request.RESPONSE.redirect(f"{target.absolute_url()}{url_param}")



from plone.protect.auto import ProtectTransform
from zope.sqlalchemy.datamanager import SessionDataManager


class EuphorieProtectTransform(ProtectTransform):
    def _registered_objects(self):
        registered = super()._registered_objects()
        app = self.request.PARENTS[-1]
        for name, conn in app._p_jar.connections.items():
            if name == "temporary":
                continue
            for resource in conn.transaction_manager.get()._resources:
                if isinstance(resource, SessionDataManager):
                    registered.extend(resource.session.dirty)
                    registered.extend(resource.session.new)
                    registered.extend(resource.session.deleted)
        return registered



from euphorie.client import CONDITIONS_VERSION
from euphorie.client import model
from euphorie.client.browser import publish
from euphorie.content.browser import upload
from Products.MailHost.MailHost import MailHost
from z3c.saconfig import Session
from zope.sendmail.mailer import SMTPMailer
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse

import sys


def addAccount(login="jane@example.com", password="Øle"):
    account = model.Account(
        loginname=login,
        password=password,
        tc_approved=CONDITIONS_VERSION,
    )
    session = Session()
    session.add(account)
    session.flush()
    return account


def addSurvey(portal, xml_survey):
    """Add a survey to the portal.

    This function requires that you are already loggin in as portal
    owner.
    """
    importer = upload.SectorImporter(portal.sectors.nl)
    sector = importer(xml_survey, None, None, None, "test import")
    survey = sector.values()[0]["test-import"]
    publisher = publish.PublishSurvey(survey, portal.REQUEST)
    publisher.publish()


def testRequest():
    """Create a new request object.

    This is based on the code in
    :py:func`Testing.makerequest.makerequest`.
    """
    environ = {"SERVER_NAME": "localhost", "SERVER_PORT": "80", "REQUEST_METHOD": "GET"}
    request = HTTPRequest(sys.stdin, environ, HTTPResponse())
    request._steps = ["Plone"]
    return request


def registerUserInClient(browser, link="Registreer"):
    browser.getLink(link).click()
    browser.getControl(name="email").value = "guest@example.com"
    browser.getControl(name="password1:utf8:ustring").value = "Guest12345#!"
    browser.getControl(name="password2:utf8:ustring").value = "Guest12345#!"
    browser.getControl(name="terms").controls[0].click()
    browser.getControl(name="register").click()
    # Screw the terms and conditions, just go to the Dashboard...
    browser.getLink("Dashboard").click()
    # # XXX Why does this not always happen??
    # if "terms-and-conditions" in browser.url:
    #     browser.getForm().submit()


class MockMailFixture:
    def __init__(self):
        self.storage = storage = []

        def send(self, *a, **kw):
            storage.append((a, kw))

        self._original_send = MailHost.send
        MailHost.send = send

    def __del__(self):
        SMTPMailer.send = self._original_send



"""
Country
-------

The main view after login, to list existing sessions, start new sessions,
delete & rename sessions

URL: https://client-oiranew.syslab.com/eu
"""

from plone.app.dexterity.behaviors.metadata import IBasic
from plone.dexterity.content import Container
from plone.supermodel import model
from zope.interface import implementer

import logging


log = logging.getLogger(__name__)


class IClientCountry(model.Schema, IBasic):
    """Country grouping in the online client."""


@implementer(IClientCountry)
class ClientCountry(Container):
    country_type = None

    # Many countries only have one language. Use it as the default
    language_mapping_by_country = {
        "bg": "bg",
        "cy": "el",
        "de": "de",
        "eu": "en",
        "fi": "fi",
        "fr": "fr",
        "gb": "en",
        "gr": "el",
        "hr": "hr",
        "hu": "hu",
        "is": "is",
        "it": "it",
        "lt": "lt",
        "lv": "lv",
        "mt": "en",
        "pt": "pt",
        "si": "sl",
        "sk": "sk",
        "cz": "cs",
    }

    @property
    def language(self):
        return self.language_mapping_by_country.get(self.id)
